"""
This module contains helper functions for the ´train´ subpackage.
"""

import sys
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np


class EarlyStopping:
    """
    EarlyStopping class for early stopping of NicheCompass training.

    Parts of the implementation are adapted from
    https://github.com/theislab/scarches/blob/cb54fa0df3255ad1576a977b17e9d77d4907ceb0/scarches/utils/monitor.py#L4
    (01.10.2022).

    Parameters
    ----------
    early_stopping_metric:
        The metric on which the early stopping criterion is calculated.
    metric_improvement_threshold:
        The minimum value which counts as metric_improvement.
    patience:
        Number of epochs which are allowed to have no metric improvement until
        the training is stopped.
    reduce_lr_on_plateau:
        If ´True´, the learning rate gets adjusted by ´lr_factor´ after a given
        number of epochs with no metric improvement.
    lr_patience:
        Number of epochs which are allowed to have no metric improvement until
        the learning rate is adjusted.
    lr_factor:
        Scaling factor for adjusting the learning rate.
     """

    def __init__(self,
                 early_stopping_metric: str = "val_global_loss",
                 metric_improvement_threshold: float = 0.,
                 patience: int = 8,
                 reduce_lr_on_plateau: bool = True,
                 lr_patience: int = 4,
                 lr_factor: float = 0.1):
        self.early_stopping_metric = early_stopping_metric
        self.metric_improvement_threshold = metric_improvement_threshold
        self.patience = patience
        self.reduce_lr_on_plateau = reduce_lr_on_plateau
        self.lr_patience = lr_patience
        self.lr_factor = lr_factor
        self.epochs = 0
        self.epochs_not_improved = 0
        self.epochs_not_improved_lr = 0
        self.current_performance = np.inf
        self.best_performance = np.inf
        self.best_performance_state = np.inf

    def step(self, current_metric: float) -> Tuple[bool, bool]:
        self.epochs += 1

        # Calculate metric improvement
        self.current_performance = current_metric
        metric_improvement = (self.best_performance -
                              self.current_performance)
        # Update best performance
        if metric_improvement > 0:
            self.best_performance = self.current_performance
        # Update epochs not improved
        if metric_improvement < self.metric_improvement_threshold:
            self.epochs_not_improved += 1
            self.epochs_not_improved_lr += 1
        else:
            self.epochs_not_improved = 0
            self.epochs_not_improved_lr = 0

        # Determine whether to continue training and whether to reduce the
        # learning rate
        if self.epochs < self.patience:
            continue_training = True
            reduce_lr = False
        elif self.epochs_not_improved >= self.patience:
            continue_training = False
            reduce_lr = False
        else:
            if self.reduce_lr_on_plateau == False:
                reduce_lr = False
            elif self.epochs_not_improved_lr >= self.lr_patience:
                reduce_lr = True
                self.epochs_not_improved_lr = 0
                print("\nReducing learning rate: metric has not improved more "
                      f"than {self.metric_improvement_threshold} in the last "
                      f"{self.lr_patience} epochs.")
            else:
                reduce_lr = False
            continue_training = True
        if not continue_training:
            print("\nStopping early: metric has not improved more than "
                  + str(self.metric_improvement_threshold) +
                  " in the last " + str(self.patience) + " epochs.")
            print("If the early stopping criterion is too strong, "
                  "please instantiate it with different parameters "
                  "in the train method.")
        return continue_training, reduce_lr

    def update_state(self, current_metric: float) -> bool:
        improved = (self.best_performance_state - current_metric) > 0
        if improved:
            self.best_performance_state = current_metric
        return improved


def print_progress(epoch: int, logs: dict, n_epochs: int):
    """
    Create message for '_print_progress_bar()' and print it out with a progress
    bar.

    Implementation is adapted from
    https://github.com/theislab/scarches/blob/master/scarches/trainers/trvae/_utils.py#L11
    (01.10.2022).

    Parameters
    ----------
    epoch:
         Current epoch.
    logs:
         Dictionary with all logs (losses & metrics).
    n_epochs:
         Total number of epochs.
    """
    # Define progress message
    message = ""
    for key in logs:
        message += f"{key:s}: {logs[key][-1]:.4f}; "
    message = message[:-2] + "\n"

    # Display progress bar
    _print_progress_bar(epoch + 1,
                        n_epochs,
                        prefix=f"Epoch {epoch + 1}/{n_epochs}",
                        suffix=message,
                        decimals=1,
                        length=20)


def _print_progress_bar(epoch: int,
                        n_epochs: int,
                        prefix: str = "",
                        suffix: str = "",
                        decimals: int = 1,
                        length: int = 100,
                        fill: str = "█"):
    """
    Print out a message with a progress bar.

    Implementation is adapted from
    https://github.com/theislab/scarches/blob/master/scarches/trainers/trvae/_utils.py#L41
    (01.10.2022).

    Parameters
    ----------
    epoch:
         Current epoch.
    n_epochs:
         Total number of epochs.
    prefix:
         String before the progress bar.
    suffix:
         String after the progress bar.
    decimals:
         Digits after comma for the percent display.
    length:
         Length of the progress bar.
    fill:
         Symbol for filling the bar.
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (
            epoch / float(n_epochs)))
    filled_len = int(length * epoch // n_epochs)
    bar = fill * filled_len + '-' * (length - filled_len)
    sys.stdout.write("\r%s |%s| %s%s %s" % (prefix, bar, percent, "%", suffix)),
    if epoch == n_epochs:
        sys.stdout.write("\n")
    sys.stdout.flush()

def _cycle_iterable(iterable):
    iterator = iter(iterable)
    while True:
        try:
            yield next(iterator)
        except StopIteration:
            iterator = iter(iterable)



from .dataloaders import initialize_dataloaders
from .dataprocessors import (edge_level_split,
                             node_level_split_mask,
                             prepare_data)
from .datareaders import (
	read_mtx,
    read_scData,
    read_multi_scData,
    concat_data
)
from .datasets import GraphAnnTorchDataset

__all__ = ["initialize_dataloaders",
           "edge_level_split",
           "node_level_split_mask",
           "prepare_data",
           "read_scData",
           "read_multi_scData",
           "concat_data",
           "GraphAnnTorchDataset"]


import numpy as np
import pandas as pd
import os
from collections import defaultdict
import networkx as nx
from scipy.spatial import distance

import scipy
from scipy.sparse import csr_matrix
from sklearn.cross_decomposition import CCA
from sklearn.neighbors import NearestNeighbors

import scanpy as sc
from anndata import AnnData

from ._utils import reindex
from .match_utils import get_initial_matching, get_refined_matching, fit_svd_on_full_data, address_matching_redundancy
from . import _utils as utils
from . import _graph as graph

## Spatial graph construction
def graph_computing(adata, k_cutoff=None, model='Radius', batch_key=None, verbose=True):
    """\
    Construct the spatial neighbor networks.
    """
    assert model in ['mu_std', 'Radius', 'KNN'], "Invalid model specified"
    if verbose:
        print('------Calculating spatial graph...')

    edgeList = []

    if batch_key is not None and len(np.unique(adata.obs[batch_key])) > 1:
        sample_names = np.unique(adata.obs[batch_key])
    else:
        sample_names = [None]

    for sample_name in sample_names:
        if sample_name is not None:
            sub_adata = adata[adata.obs[batch_key] == sample_name].copy()
        else:
            sub_adata = adata.copy()

        coor = pd.DataFrame(sub_adata.obsm['spatial'])
        coor.index = sub_adata.obs.index
        coor.columns = ['imagerow', 'imagecol']

        if model == 'mu_std':
            distMat = distance.cdist(coor, coor, 'euclidean')  # compute all distances at once
            for node_idx in range(sub_adata.shape[0]):
                nearest_indices = np.argsort(distMat[node_idx])[:k_cutoff + 1]
                nearest_dists = distMat[node_idx, nearest_indices[1:k_cutoff + 1]]
                boundary = np.mean(nearest_dists) + np.std(nearest_dists)
                for j, dist in zip(nearest_indices[1:], nearest_dists):
                    weight = 1.0 if dist <= boundary else 0.0
                    edgeList.append((node_idx, j, weight))

        elif model == 'Radius':
            nbrs = NearestNeighbors(radius=k_cutoff).fit(coor)
            distances, indices = nbrs.radius_neighbors(coor, return_distance=True)
            for it in range(indices.shape[0]):
                edgeList.extend(list(zip([it] * indices[it].shape[0], indices[it], distances[it])))

        elif model == 'KNN':
            nbrs = NearestNeighbors(n_neighbors=k_cutoff + 1).fit(coor)
            distances, indices = nbrs.kneighbors(coor)
            for it in range(indices.shape[0]):
                edgeList.extend(list(zip([it] * indices.shape[1], indices[it, :], distances[it, :])))

    return edgeList

def edgeList2edgeDict(edgeList, node_size):
    graphdict = defaultdict(list)
    for start, end, _ in edgeList:
        graphdict[start].append(end)
    return {node: graphdict[node] for node in range(node_size)}  # ensure all nodes are present

def graph_construction(adata, mode, k, batch_key, verbose=True):
    edge_list = graph_computing(adata, k_cutoff=k, model=mode, batch_key=batch_key, verbose=verbose)
    graph_dict = edgeList2edgeDict(edge_list, adata.shape[0])
    adj_org = nx.adjacency_matrix(nx.from_dict_of_lists(graph_dict))
    return adj_org



def find_initial_pivots(
        arr1, arr2,
        svd_components1=None, svd_components2=None,
        randomized_svd=False, svd_runs=1,
        verbose=True
):
    """
    Perform initial matching.

    Parameters
    ----------
    svd_components1: None or int, default=None
        If not None, perform SVD to reduce the dimension of self.shared_arr1.
    svd_components2: None or int, default=None
        If not None, perform SVD to reduce the dimension of self.shared_arr2.
    randomized_svd: bool, default=False
        Whether to use randomized SVD.
    svd_runs: int, default=1
        Perform multiple runs of SVD and the one with lowest Frobenious reconstruction error is selected.
    verbose: bool, default=True
        Whether to print the progress.

    Returns
    -------
    None
    """
    init_matching = []
    init_matching.append(
        get_initial_matching(
            arr1=arr1,
            arr2=arr2,
            randomized_svd=randomized_svd,
            svd_runs=svd_runs,
            svd_components1=svd_components1,
            svd_components2=svd_components2,
            verbose=False
        )
    )

    if verbose:
        print('Init_matching done!', flush=True)
    return init_matching


def refine_pivots(
        init_matching, arr1, arr2,
        svd_components1=None, svd_components2=None,
        cca_components=None,
        filter_prop=0,
        n_iters=1,
        randomized_svd=False, svd_runs=1,
        cca_max_iter=2000,
        verbose=True
):
    refined_matching = []
    refined_matching.append(
        get_refined_matching(
            init_matching=init_matching,
            arr1=arr1,
            arr2=arr2,
            randomized_svd=randomized_svd,
            svd_runs=svd_runs,
            svd_components1=svd_components1,
            svd_components2=svd_components2,
            n_iters=n_iters,
            filter_prop=filter_prop,
            cca_components=cca_components,
            cca_max_iter=cca_max_iter,
            verbose=False
        )
    )

    if verbose:
        print('Refined_matching done!', flush=True)
    return refined_matching


def filter_bad_matches(arr1, arr2, refined_matching=None,
                       propagated_matching=None, target='pivot',
                       filter_prop=0., randomized_svd=False, svd_runs=1,
                       svd_components1=None, svd_components2=None,
                       cca_components=20,
                       cca_max_iter=2000,
                       verbose=True):
    arr1_raw = arr1
    arr2_raw = arr2
    # conduct filtering
    n_remaining = 0
    if verbose:
        print('Begin filtering...', flush=True)
    if target == 'pivot':
        matching_to_be_filtered = refined_matching
    elif target == 'propagated':
        matching_to_be_filtered = propagated_matching
    else:
        raise ValueError('target must be in {\'pivot\', \'propagated\'}.')

    remaining_indices_after_filtering = []
    idx2_to_indices1 = defaultdict(set)
    # record the locations that survive the filtering
    rows, cols, vals = matching_to_be_filtered[0]
    # anything with val <= thresh will be retained
    thresh = np.quantile(vals, 1 - filter_prop)
    remaining_indices_after_filtering.append([i for i in range(len(vals)) if vals[i] <= thresh])
    n_remaining += len(remaining_indices_after_filtering[0])
    for i in remaining_indices_after_filtering[0]:
        idx1, idx2 = rows[i], cols[i]
        idx2_to_indices1[idx2].add(idx1)

    if target == 'pivot':
        remaining_indices_in_refined_matching = remaining_indices_after_filtering[0]
    else:
        remaining_indices_in_propagated_matching = remaining_indices_after_filtering[0]

    if verbose:
        print('{}/{} pairs of matched cells remain after the filtering.'.format(
            n_remaining, np.sum([len(per_batch_matching[0]) for per_batch_matching in matching_to_be_filtered])
        ))

    # convert idx2_to_indices1 to a dict of lists for ease of later usage
    idx2_to_indices1 = {idx2: sorted(indices1) for idx2, indices1 in idx2_to_indices1.items()}

    # fit CCA on pivots
    if verbose:
        print('Fitting CCA on pivots...', flush=True)
    rotation_before_cca_on_active_arr1, rotation_before_cca_on_active_arr2 = fit_svd_on_full_data(arr1, arr2, randomized_svd=randomized_svd,
                                                                                                  svd_runs=svd_runs, svd_components1=svd_components1,
                                                                                                  svd_components2=svd_components2)
    arr1_list, arr2_list = [], []
    for idx2, indices1 in idx2_to_indices1.items():
        arr1_list.append(arr1[indices1[0], :].toarray().squeeze())  # np.mean(self.active_arr1[indices1, :], axis=0)
        arr2_list.append(arr2[idx2, :].toarray().squeeze())
    arr1 = rotation_before_cca_on_active_arr1(np.array(arr1_list))
    arr2 = rotation_before_cca_on_active_arr2(np.array(arr2_list))

    cca_on_pivots = CCA(n_components=cca_components, max_iter=cca_max_iter)
    cca_on_pivots.fit(arr1, arr2)

    if verbose:
        print('Scoring matched pairs...', flush=True)
    # transform the whole dataset
    arr1_raw = rotation_before_cca_on_active_arr1(arr1_raw)
    arr2_raw = rotation_before_cca_on_active_arr2(arr2_raw)
    arr1, arr2 = cca_on_pivots.transform(arr1_raw, arr2_raw)
    arr1 = utils.center_scale(arr1)
    arr2 = utils.center_scale(arr2)

    # add distances to idx2_to_indices1
    all_indices1, all_indices2 = [], []
    for idx2, indices1 in idx2_to_indices1.items():
        for idx1 in indices1:
            all_indices1.append(idx1)
            all_indices2.append(idx2)
    # compute all possible pairs of distances
    pearson_correlations = utils.pearson_correlation(arr1[all_indices1, :], arr2[all_indices2, :])

    cnt = 0
    for idx2, indices1 in idx2_to_indices1.items():
        indices_and_scores = []
        for idx1 in indices1:
            indices_and_scores.append((idx1, pearson_correlations[cnt]))
            cnt += 1
        idx2_to_indices1[idx2] = indices_and_scores

    if target == 'pivot':
        pivot2_to_pivots1 = idx2_to_indices1
        return pivot2_to_pivots1, remaining_indices_in_refined_matching
    else:
        propidx2_to_propindices1 = idx2_to_indices1
        return propidx2_to_propindices1, remaining_indices_in_propagated_matching


def propagate(refined_matching=None, remaining_indices_in_refined_matching=None,
              arr1=None, arr2=None, svd_components1=None, svd_components2=None,
              metric='euclidean', randomized_svd=False, svd_runs=1, verbose=True):
    """
    For indices not in pivots, find their matches by nearest neighbor search.

    Parameters
    ----------
    svd_components1: None or int, default=None
        If not None, perform SVD to reduce the dimension of self.active_arr1
        before doing internal nearest neighbor search.
    svd_components2: None or int, default=None
        If not None, perform SVD to reduce the dimension of self.active_arr1
        before doing internal nearest neighbor search.
    wt1: float, default=0.7
        Weight to put on raw data of self.active_arr1 when doing smoothing.
    wt2: float, default=0.7
        Weight to put on raw data of self.active_arr2 when doing smoothing.
    metric: string, default='correlation'
        The metric to use in nearest neighbor search.
    randomized_svd: bool, default=False
        Whether to perform randomized SVD.
    svd_runs: int, default=1
        Perform multiple runs of SVD and the one with lowest Frobenious reconstruction error is selected.
    verbose: bool, default=True
        Whether to print the progress.

    Returns
    -------
    None
    """
    propagated_matching = []
    curr_propagated_matching = [[], [], []]
    curr_refined_matching = refined_matching
    # get good pivot indices that survived pivot filtering
    existing_indices = curr_refined_matching[0][remaining_indices_in_refined_matching]
    good_indices1 = curr_refined_matching[0][existing_indices]
    good_indices2 = curr_refined_matching[1][existing_indices]

    # get arrays that were used when doing refined matching
    # get remaining indices
    # propagation will only be done for those indices
    good_indices1_set = set(good_indices1)
    remaining_indices1 = [i for i in range(arr1.shape[0]) if i not in good_indices1_set]
    good_indices2_set = set(good_indices2)
    remaining_indices2 = [i for i in range(arr2.shape[0]) if i not in good_indices2_set]

    # propagate for remaining indices in arr1
    if len(remaining_indices1) > 0:
        # get 1-nearest-neighbors and the corresponding distances
        remaining_indices1_nns, remaining_indices1_nn_dists = graph.get_nearest_neighbors(
            query_arr=arr1[remaining_indices1, :],
            target_arr=arr1[good_indices1, :],
            svd_components=svd_components1,
            randomized_svd=randomized_svd,
            svd_runs=svd_runs,
            metric=metric
        )
        matched_indices2 = good_indices2[remaining_indices1_nns]
        curr_propagated_matching[0].extend(remaining_indices1)
        curr_propagated_matching[1].extend(matched_indices2)
        curr_propagated_matching[2].extend(remaining_indices1_nn_dists)

    # propagate for remaining indices in arr2
    if len(remaining_indices2) > 0:
        # get 1-nearest-neighbors and the corresponding distances
        remaining_indices2_nns, remaining_indices2_nn_dists = graph.get_nearest_neighbors(
            query_arr=arr2[remaining_indices2, :],
            target_arr=arr2[good_indices2, :],
            svd_components=svd_components2,
            randomized_svd=randomized_svd,
            svd_runs=svd_runs,
            metric=metric
        )
        matched_indices1 = good_indices1[remaining_indices2_nns]
        curr_propagated_matching[0].extend(matched_indices1)
        curr_propagated_matching[1].extend(remaining_indices2)
        curr_propagated_matching[2].extend(remaining_indices2_nn_dists)

    propagated_matching.append(
        (np.array(curr_propagated_matching[0]),
         np.array(curr_propagated_matching[1]),
         np.array(curr_propagated_matching[2]))
    )

    if verbose:
        print('Done!', flush=True)

    return propagated_matching

def get_matching(pivot2_to_pivots1=None, propidx2_to_propindices1=None, arr1=None, arr2=None, direction='both', target='pivot'):
    """
    Return a copy of the desired matching.

    Parameters
    ----------
    order: None or (1, 2) or (1, 2), default=None
        If (1, 2), then every cell in target arr1 has at least one match
        if (2, 1), then does the other way around,
        if None, then every cell in target arr1 and every cell in target arr2 both have at least one match
    target: 'pivot' or 'full_data'
        If 'pivot', then only return matching on pivots, else return matching on all the data.

    Returns
    -------
    A matching of format dict or list.
    """
    if target not in {'pivot', 'full_data'}:
        raise ValueError('mode must be in {\'pivot_only\', \'full_data\'}.')

    res = [[], [], []]
    for idx2, indices1_and_scores in pivot2_to_pivots1.items():
        for idx1, score in indices1_and_scores:
            res[0].append(idx1)
            res[1].append(idx2)
            res[2].append(score)
    if target == 'pivot':
        return res
    elif target == 'full_data':
        if direction == 'left':
            # add propagated matching for non-pivot cells in the first dataset
            existing_indices1 = np.unique(res[0])
            remaining_indices1 = [i for i in range(arr1.shape[0]) if i not in existing_indices1]
            propagated_idx1_to_indices2 = defaultdict(list)
            for idx2, indices1_and_scores in propidx2_to_propindices1.items():
                for idx1, score in indices1_and_scores:
                    propagated_idx1_to_indices2[idx1].append((idx2, score))
            for idx1 in remaining_indices1:
                if idx1 in propagated_idx1_to_indices2:
                    for idx2, score in propagated_idx1_to_indices2[idx1]:
                        res[0].append(idx1)
                        res[1].append(idx2)
                        res[2].append(score)
        elif direction == 'right':
            # add propagated matching for non-pivot cells in the second dataset
            existing_indices2 = np.unique(res[1])
            remaining_indices2 = [i for i in range(arr2.shape[0]) if i not in existing_indices2]
            for idx2 in remaining_indices2:
                if idx2 in propidx2_to_propindices1:
                    for idx1, score in propidx2_to_propindices1[idx2]:
                        res[0].append(idx1)
                        res[1].append(idx2)
                        res[2].append(score)
        elif direction == 'both':
            # first do order (1, 2) and then do order (2, 1)
            # add propagated matching for non-pivot cells in the first dataset
            existing_indices1 = np.unique(res[0])
            remaining_indices1 = [i for i in range(arr1.shape[0]) if i not in existing_indices1]
            propagated_idx1_to_indices2 = defaultdict(list)
            for idx2, indices1_and_scores in propidx2_to_propindices1.items():
                for idx1, score in indices1_and_scores:
                    propagated_idx1_to_indices2[idx1].append((idx2, score))
            for idx1 in remaining_indices1:
                if idx1 in propagated_idx1_to_indices2:
                    for idx2, score in propagated_idx1_to_indices2[idx1]:
                        res[0].append(idx1)
                        res[1].append(idx2)
                        res[2].append(score)

            # add propagated matching for non-pivot cells in the second dataset
            existing_indices2 = np.unique(res[1])
            remaining_indices2 = [i for i in range(arr2.shape[0]) if i not in existing_indices2]
            for idx2 in remaining_indices2:
                if idx2 in propidx2_to_propindices1:
                    for idx1, score in propidx2_to_propindices1[idx2]:
                        res[0].append(idx1)
                        res[1].append(idx2)
                        res[2].append(score)

        else:
            raise NotImplementedError('direction must be `both` or `left` or `right`.')
    else:
        raise NotImplementedError('target must be in {\'pivot\', \'full_data\'}.')

    return address_matching_redundancy(matching=res, direction=direction)

## Main function
def create_adj(rna, atac, rna_adata_shared, atac_adata_shared, data_type='Paired', rna_n_top_features=3000, atac_n_top_features=10000,
               batch_key=None, svd_components1=30, svd_components2=30,
               cca_components=20, cca_max_iter=2000, randomized_svd=False,
               filter_prop_initial=0, filter_prop_refined=0.3, filter_prop_propagated=0,
               n_iters=1, svd_runs=1, verbose=True):

    # Finding initial pivots on all genes or peaks
    initial_pivots = find_initial_pivots(rna_adata_shared.X, atac_adata_shared.X, svd_components1=svd_components1, svd_components2=svd_components2,
                                         randomized_svd=randomized_svd, svd_runs=svd_runs, verbose=verbose)

    # Finding variable features for RNA adata
    if type(rna_n_top_features) == int:
        if rna_n_top_features > len(rna.var_names):
            rna_n_top_features = len(rna.var_names)
        if batch_key is not None:
            sc.pp.highly_variable_genes(rna, n_top_genes=rna_n_top_features, batch_key=batch_key)
        else:
            sc.pp.highly_variable_genes(rna, n_top_genes=rna_n_top_features)
            # sc.pp.highly_variable_genes(adata, min_mean=0.0125, max_mean=3, min_disp=0.5)
        rna = rna[:, rna.var.highly_variable].copy()
    elif type(rna_n_top_features) != int:
        if isinstance(rna_n_top_features, str):
            if os.path.isfile(rna_n_top_features):
                rna_n_top_features = np.loadtxt(rna_n_top_features, dtype=str)
        rna = reindex(rna, rna_n_top_features)

    # Finding variable features for ATAC adata
    if type(atac_n_top_features) == int and atac_n_top_features > 0 and atac_n_top_features < atac.shape[1]:
        if batch_key is not None:
            sc.pp.highly_variable_genes(atac, n_top_genes=atac_n_top_features,
                                        batch_key=batch_key,
                                        inplace=False, subset=True)
        else:
            sc.pp.highly_variable_genes(atac, n_top_genes=atac_n_top_features,
                                        inplace=False, subset=True)
    elif type(atac_n_top_features) != int:
        if isinstance(atac_n_top_features, str):
            if os.path.isfile(atac_n_top_features):
                atac_n_top_features = np.loadtxt(atac_n_top_features, dtype=str)
        atac = reindex(atac, atac_n_top_features)

    # Refining pivots
    fine_pivots = refine_pivots(initial_pivots[0], rna.X, atac.X, svd_components1=svd_components1, svd_components2=svd_components2,
                                cca_components=cca_components, filter_prop=filter_prop_initial, n_iters=n_iters,
                                randomized_svd=randomized_svd, svd_runs=svd_runs, cca_max_iter=cca_max_iter, verbose=verbose)

    # Filtering bad matches
    print('Filter_bad_matches on pivots matching...')
    pivot2_to_pivots1, remaining_indices_in_refined_matching = filter_bad_matches(
        rna.X, atac.X, refined_matching=fine_pivots, target='pivot', filter_prop=filter_prop_refined,
        randomized_svd=randomized_svd, svd_runs=svd_runs, svd_components1=svd_components1, svd_components2=svd_components2,
        cca_components=cca_components, cca_max_iter=cca_max_iter, verbose=verbose)

    # Propagating matching
    propagated_matching = propagate(refined_matching=fine_pivots[0], remaining_indices_in_refined_matching=remaining_indices_in_refined_matching,
                                    arr1=rna.X, arr2=atac.X, svd_components1=svd_components1, svd_components2=svd_components2, metric='euclidean',
                                    randomized_svd=randomized_svd, svd_runs=svd_runs, verbose=verbose)

    # Final filtering of bad matches
    print('Filter_bad_matches on propagated matching...')
    propidx2_to_propindices1, remaining_indices_in_propagated_matching = filter_bad_matches(
        rna.X, atac.X, propagated_matching=propagated_matching, target='propagated', filter_prop=filter_prop_propagated,
        randomized_svd=randomized_svd, svd_runs=svd_runs, svd_components1=svd_components1, svd_components2=svd_components2,
        cca_components=cca_components, cca_max_iter=cca_max_iter, verbose=verbose)

    # Generating final matching
    final_cellmatching = get_matching(pivot2_to_pivots1, propidx2_to_propindices1, arr1=rna.X, arr2=atac.X, direction='both', target='full_data')
    df = pd.DataFrame(list(zip(final_cellmatching[0], final_cellmatching[1], final_cellmatching[2])),
                      columns=['mod1_indx', 'mod2_indx', 'score'])

    # Preparing pivot table and converting to sparse matrix format
    if data_type == 'Paired':
        all_mod1_indx = pd.Index(range(rna.shape[0]))
        all_mod2_indx = pd.Index(range(atac.shape[0]))
        pivot_df = df.pivot(index='mod1_indx', columns='mod2_indx', values='score').reindex(index=all_mod1_indx, columns=all_mod2_indx).fillna(0.)
        sparse_matrix = csr_matrix(pivot_df)
    elif data_type == 'UnPaired' or data_type == 'multi-modal':
        # df2 = df.copy()
        all_mod1_indx = pd.Index(range(rna.shape[0]))
        all_mod2_indx = pd.Index(range(atac.shape[0]))
        # all_mod1_indx = pd.Index(range((rna.shape[0] + atac.shape[0])))
        # all_mod2_indx = pd.Index(range((rna.shape[0] + atac.shape[0])))
        # df['mod1_indx'] += atac.shape[0]  # 将RNA索引偏移
        # df['mod2_indx'] += rna.shape[0]  # 将ATAC索引偏移
        pivot_df = df.pivot(index='mod1_indx', columns='mod2_indx', values='score').reindex(index=all_mod1_indx, columns=all_mod2_indx).fillna(0.)

        # df2['mod1_indx'] += atac.shape[0]  # 将RNA索引偏移
        # df2['mod2_indx'] += rna.shape[0]  # 将ATAC索引偏移
        # pivot_df2 = df2.pivot(index='mod1_indx', columns='mod2_indx', values='score').reindex(index=all_mod1_indx, columns=all_mod2_indx).fillna(0.)
        # combined_pivot_df = pivot_df.add(pivot_df2, fill_value=0)

        ## 增加对角线的weight
        # 设置主对角线的值为1
        # np.fill_diagonal(combined_pivot_df.values, 1)

        # compute all possible pairs of distances for rna and atac
        # pearson_correlations_rna = utils.pearson_correlation(rna.X, rna.X)
        # pearson_correlations_atac = utils.pearson_correlation(atac.X, atac.X)
        # # Generate row and column indices
        # row_indices = np.arange(len(pearson_correlations_rna)).astype(int)
        # col_indices = np.arange(len(correlation_values)).astype(int)
        #
        # # Construct the output array with row, col, value
        # output_array = np.column_stack((row_indices, col_indices, pearson_correlations_rna))

        sparse_matrix = csr_matrix(pivot_df)

    return sparse_matrix










"""
This module contains all loss functions used by the Garfield module.
"""

from typing import List, Literal, Optional

import math
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

def compute_omics_recon_mse_loss(recon_x, x):
    """Computes MSE loss between reconstructed data and ground truth data.

       Parameters
       ----------
       recon_x: torch.Tensor
            Torch Tensor of reconstructed data
       x: torch.Tensor
            Torch Tensor of ground truth data

       Returns
       -------
       MSE loss value
    """
    mse_loss = F.mse_loss(recon_x, x) # , reduction='sum'
    return mse_loss

def compute_edge_recon_loss(
        edge_recon_logits: torch.Tensor,
        edge_recon_labels: torch.Tensor,
        edge_incl: Optional[torch.Tensor]=None
        ) -> torch.Tensor:
    """
    Compute edge reconstruction weighted binary cross entropy loss with logits
    using ground truth edge labels and predicted edge logits.

    Parameters
    ----------
    edge_recon_logits:
        Predicted edge reconstruction logits for both positive and negative
        sampled edges (dim: 2 * ´edge_batch_size´).
    edge_recon_labels:
        Edge ground truth labels for both positive and negative sampled edges
        (dim: 2 * ´edge_batch_size´).
    edge_incl:
        Boolean mask which indicates edges to be included in the edge recon loss
        (dim: 2 * ´edge_batch_size´). If ´None´, includes all edges.

    Returns
    ----------
    edge_recon_loss:
        Weighted binary cross entropy loss between edge labels and predicted
        edge probabilities (calculated from logits for numerical stability in
        backpropagation).
    """
    if edge_incl is not None:
        # Remove edges whose node pair has different categories in categorical
        # covariates for which no cross-category edges are present
        edge_recon_logits = edge_recon_logits[edge_incl]
        edge_recon_labels = edge_recon_labels[edge_incl]

    # Determine weighting of positive examples
    pos_labels = (edge_recon_labels == 1.).sum(dim=0)
    neg_labels = (edge_recon_labels == 0.).sum(dim=0)
    pos_weight = neg_labels / pos_labels

    # Compute weighted bce loss from logits for numerical stability
    edge_recon_loss = F.binary_cross_entropy_with_logits(edge_recon_logits,
                                                         edge_recon_labels,
                                                         pos_weight=pos_weight)
    return edge_recon_loss


def compute_kl_reg_loss(mu: torch.Tensor,
                        logstd: torch.Tensor) -> torch.Tensor:
    """
    Compute Kullback-Leibler divergence as per Kingma, D. P. & Welling, M.
    Auto-Encoding Variational Bayes. arXiv [stat.ML] (2013). Equation (10).
    This will encourage encodings to distribute evenly around the center of
    a continuous and complete latent space, producing similar (for points close
    in latent space) and meaningful content after decoding.

    For detailed derivation, see
    https://stats.stackexchange.com/questions/318748/deriving-the-kl-divergence-loss-for-vaes.

    Parameters
    ----------
    mu:
        Expected values of the normal latent distribution of each node (dim:
        n_nodes_current_batch, n_gps).
    logstd:
        Log standard deviations of the normal latent distribution of each node
        (dim: n_nodes_current_batch, n_gps).

    Returns
    ----------
    kl_reg_loss:
        Kullback-Leibler divergence.
    """
    # Sum over n_gps and mean over n_nodes_current_batch
    kl_reg_loss = -0.5 * torch.mean(
        torch.sum(1 + 2 * logstd - mu ** 2 - torch.exp(logstd) ** 2, 1))
    return kl_reg_loss


## contrastive loss
def compute_contrastive_instanceloss(z_i, z_j, temperature):
    """
    Compute the contrastive loss given two batches of feature vectors z_i and z_j.

    Parameters:
    z_i (Tensor): Feature vectors from the first view.
    z_j (Tensor): Feature vectors from the second view.
    temperature (float): Temperature parameter to scale the dot products.

    Returns:
    loss (Tensor): The computed contrastive loss.
    """

    # Initialize Cross Entropy Loss
    criterion = nn.CrossEntropyLoss(reduction="sum")
    def mask_correlated_samples(batch_size):
        """
        Creates a mask to zero out correlations between the same samples.

        Parameters:
        batch_size (int): The number of samples in one batch.

        Returns:
        mask (Tensor): A mask of shape (2*batch_size, 2*batch_size) where correlated samples have zero value.
        """
        N = 2 * batch_size
        mask = torch.ones((N, N))
        mask = mask.fill_diagonal_(0)
        for i in range(batch_size):
            mask[i, batch_size + i] = 0
            mask[batch_size + i, i] = 0
        mask = mask.bool()
        return mask

    # Compute batch size and mask for correlated samples
    batch_size = z_i.size(0)
    mask = mask_correlated_samples(batch_size)
    N = 2 * batch_size
    # Concatenate feature vectors
    z = torch.cat((z_i, z_j), dim=0)
    # Compute similarity matrix
    sim = torch.matmul(z, z.T) / temperature
    # Extract positive samples (diagonal elements)
    sim_i_j = torch.diag(sim, batch_size)
    sim_j_i = torch.diag(sim, -batch_size)
    positive_samples = torch.cat((sim_i_j, sim_j_i), dim=0).reshape(N, 1)
    # Extract negative samples
    negative_samples = sim[mask].reshape(N, -1)
    # Create labels (positive samples are labeled as 0)
    labels = torch.zeros(N).to(positive_samples.device).long()
    # Concatenate positive and negative samples
    logits = torch.cat((positive_samples, negative_samples), dim=1)
    # Compute loss
    loss = criterion(logits, labels)
    loss /= N

    return loss



def compute_contrastive_clusterloss(c_i, c_j, class_num, temperature):
    """
    Cluster loss function.

    Args:
        c_i (torch.Tensor): First set of cluster probabilities.
        c_j (torch.Tensor): Second set of cluster probabilities.
        class_num (int): Number of classes.
        temperature (float): Temperature scaling factor.
        device (torch.device): The device to perform computations on.

    Returns:
        torch.Tensor: The computed loss value.
    """

    # Create the mask for correlated clusters
    def mask_correlated_clusters(class_num):
        N = 2 * class_num
        mask = torch.ones((N, N))
        mask = mask.fill_diagonal_(0)
        for i in range(class_num):
            mask[i, class_num + i] = 0
            mask[class_num + i, i] = 0
        mask = mask.bool()
        return mask

    # Initialize necessary components
    criterion = nn.CrossEntropyLoss(reduction="sum")
    similarity_f = nn.CosineSimilarity(dim=2)

    # Compute negative entropy loss for c_i and c_j
    p_i = c_i.sum(0).view(-1)
    p_i /= p_i.sum()
    ne_i = math.log(p_i.size(0)) + (p_i * torch.log(p_i)).sum()

    p_j = c_j.sum(0).view(-1)
    p_j /= p_j.sum()
    ne_j = math.log(p_j.size(0)) + (p_j * torch.log(p_j)).sum()

    ne_loss = ne_i + ne_j

    # Concatenate c_i and c_j
    c_i = c_i.t()
    c_j = c_j.t()
    N = 2 * class_num
    c = torch.cat((c_i, c_j), dim=0)

    # Compute similarity
    sim = similarity_f(c.unsqueeze(1), c.unsqueeze(0)) / temperature
    sim_i_j = torch.diag(sim, class_num)
    sim_j_i = torch.diag(sim, -class_num)

    # Select positive and negative clusters
    positive_clusters = torch.cat((sim_i_j, sim_j_i), dim=0).reshape(N, 1)
    mask = mask_correlated_clusters(class_num)
    negative_clusters = sim[mask].reshape(N, -1)

    # Compute loss
    labels = torch.zeros(N).to(positive_clusters.device).long()
    logits = torch.cat((positive_clusters, negative_clusters), dim=1)
    loss = criterion(logits, labels)
    loss /= N

    return loss + ne_loss


### mmd function
def pairwise_distance(x, y):
    x = x.view(x.shape[0], x.shape[1], 1)
    y = torch.transpose(y, 0, 1)
    output = torch.sum((x - y) ** 2, 1)
    output = torch.transpose(output, 0, 1)

    return output

def gaussian_kernel_matrix(x, y, alphas):
    """Computes multiscale-RBF kernel between x and y.

       Parameters
       ----------
       x: torch.Tensor
            Tensor with shape [batch_size, z_dim].
       y: torch.Tensor
            Tensor with shape [batch_size, z_dim].
       alphas: Tensor

       Returns
       -------
       Returns the computed multiscale-RBF kernel between x and y.
    """

    dist = pairwise_distance(x, y).contiguous()
    dist_ = dist.view(1, -1)

    alphas = alphas.view(alphas.shape[0], 1)
    beta = 1. / (2. * alphas)

    s = torch.matmul(beta, dist_)

    return torch.sum(torch.exp(-s), 0).view_as(dist)

def compute_omics_recon_mmd_loss(source_features, target_features):
    """Initializes Maximum Mean Discrepancy(MMD) between source_features and target_features.

       - Gretton, Arthur, et al. "A Kernel Two-Sample Test". 2012.

       Parameters
       ----------
       source_features: torch.Tensor
            Tensor with shape [batch_size, z_dim]
       target_features: torch.Tensor
            Tensor with shape [batch_size, z_dim]

       Returns
       -------
       Returns the computed MMD between x and y.
    """
    alphas = [
        1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1, 5, 10, 15, 20, 25, 30, 35, 100,
        1e3, 1e4, 1e5, 1e6
    ]
    alphas = Variable(torch.FloatTensor(alphas)).to(device=source_features.device)

    cost = torch.mean(gaussian_kernel_matrix(source_features, source_features, alphas))
    cost += torch.mean(gaussian_kernel_matrix(target_features, target_features, alphas))
    cost -= 2 * torch.mean(gaussian_kernel_matrix(source_features, target_features, alphas))

    return cost




import numpy as np

import torch
import torch.nn as nn

# mask feature function
def drop_feature(x, drop_prob):
    drop_mask = torch.empty(
        (x.size(1), ),
        dtype=torch.float32,
        device=x.device).uniform_(0, 1) < drop_prob
    x = x.clone()
    x[:, drop_mask] = 0

    return x

# Domain-specific Batch Normalization（领域特定的批次归一化）是一种调整神经网络中批次归一化层以适应不同输入数据源或领域的技术。
class DSBatchNorm(nn.Module):
    """
    Domain-specific Batch Normalization
    """

    def __init__(self, num_features, n_domain, eps=1e-5, momentum=0.1):
        """
        Parameters
        ----------
        num_features
            dimension of the features
        n_domain
            domain number
        """
        super().__init__()
        self.device = torch.device("cuda:1" if torch.cuda.is_available() else "cpu")
        self.n_domain = n_domain
        self.num_features = num_features
        self.bns = nn.ModuleList(
            [nn.BatchNorm1d(num_features, eps=eps, momentum=momentum) for i in range(n_domain)]).to(self.device)

    def reset_running_stats(self):
        for bn in self.bns:
            bn.reset_running_stats()

    def reset_parameters(self):
        for bn in self.bns:
            bn.reset_parameters()

    def _check_input_dim(self, input):
        raise NotImplementedError

    def forward(self, x, y):
        device = x.device  # 获取输入数据的设备
        out = torch.zeros(x.size(0), self.num_features, device=device)  # , requires_grad=False)
        for i in range(self.n_domain):
            indices = np.where(y.cpu().numpy() == i)[0]
            if len(indices) > 1:
                out[indices] = self.bns[i](x[indices])
            elif len(indices) == 1:
                self.bns[i].training = False
                out[indices] = self.bns[i](x[indices])
                self.bns[i].training = True
        return out


def compute_cosine_similarity(tensor1: torch.Tensor,
                              tensor2: torch.Tensor,
                              eps: float = 1e-8) -> torch.Tensor:
    """
    Compute the element-wise cosine similarity between two 2D tensors.

    Parameters
    ----------
    tensor1:
        First tensor for element-wise cosine similarity computation (dim: n_obs
        x n_features).
    tensor2:
        Second tensor for element-wise cosine similarity computation (dim: n_obs
        x n_features).

    Returns
    ----------
    cosine_sim:
        Result tensor that contains the computed element-wise cosine
        similarities (dim: n_obs).
    """
    tensor1_norm = tensor1.norm(dim=1)[:, None]
    tensor2_norm = tensor2.norm(dim=1)[:, None]
    tensor1_normalized = tensor1 / torch.max(
        tensor1_norm, eps * torch.ones_like(tensor1_norm))
    tensor2_normalized = tensor2 / torch.max(
        tensor2_norm, eps * torch.ones_like(tensor2_norm))
    cosine_sim = torch.mul(tensor1_normalized, tensor2_normalized).sum(1)
    return cosine_sim



import os
import scanpy as sc
import muon as mu
import anndata as ad

from anndata import AnnData
from muon import MuData
import episcanpy as epi

import numpy as np
import pandas as pd
import scipy
from scipy.sparse import csr_matrix
from scipy.sparse import issparse, csr
from sklearn.preprocessing import MaxAbsScaler

# from ._pca import select_pcs_features
from ._utils import reindex, gene_scores, TFIDF_LSI, clr_normalize_each_cell
from .adj_construction import create_adj, graph_construction

CHUNK_SIZE = 20000

def batch_scale(adata, use_rep='X', batch_key='batch', chunk_size=CHUNK_SIZE):
    """
    Batch-specific scale data

    Parameters
    ----------
    adata
        AnnData
    use_rep
        use '.X' or '.obsm'
    chunk_size
        chunk large data into small chunks

    """
    for b in adata.obs[batch_key].unique():
        idx = np.where(adata.obs[batch_key] == b)[0]
        if use_rep == 'X':
            scaler = MaxAbsScaler(copy=False).fit(adata.X[idx])
            for i in range(len(idx) // chunk_size + 1):
                adata.X[idx[i * chunk_size:(i + 1) * chunk_size]] = scaler.transform(
                    adata.X[idx[i * chunk_size:(i + 1) * chunk_size]])
        else:
            scaler = MaxAbsScaler(copy=False).fit(adata.obsm[use_rep][idx])
            for i in range(len(idx) // chunk_size + 1):
                adata.obsm[use_rep][idx[i * chunk_size:(i + 1) * chunk_size]] = scaler.transform(
                    adata.obsm[use_rep][idx[i * chunk_size:(i + 1) * chunk_size]])

    return adata


def preprocessing_rna(
        adata: AnnData,
        min_features: int = 600,
        min_cells: int = 3,
        target_sum: int = 10000,
        used_hvgs: bool = True,
        used_pca_graph: bool = True,
        rna_n_top_features=2000,  # or gene list
        n_components: int = 50,
        n: int = 15,
        batch_key: str = 'batch',
        metric: str = 'euclidean',
        svd_solver: str = 'arpack',
        keep_mt: bool = False,
        backed: bool = False
):
    """
    Preprocessing single-cell RNA-seq data

    Parameters
    ----------
    adata
        An AnnData matrice of shape n_obs x n_vars. Rows correspond to cells and columns to genes.
    min_features
        Filtered out cells that are detected in less than n genes. Default: 600.
    min_cells
        Filtered out genes that are detected in less than n cells. Default: 3.
    target_sum
        After normalization, each cell has a total count equal to target_sum. If None, total count of each cell equal to the median of total counts for cells before normalization.
    rna_n_top_features
        Number of highly-variable genes to keep. Default: 2000.

    Return
    -------
    The AnnData object after preprocessing.
    """
    if min_features is None: min_features = 600
    if rna_n_top_features is None: rna_n_top_features = 2000
    if target_sum is None: target_sum = 10000

    # Preprocessing
    # if not issparse(adata.X):
    if type(adata.X) != csr.csr_matrix and (not backed) and (not adata.isbacked):
        adata.X = scipy.sparse.csr_matrix(adata.X)

    if not keep_mt:
        # Filtering out MT genes
        adata = adata[:, [gene for gene in adata.var_names
                          if not str(gene).startswith(tuple(['ERCC', 'MT-', 'mt-']))]]

    # Filtering cells
    sc.pp.filter_cells(adata, min_genes=min_features)

    # Filtering features
    sc.pp.filter_genes(adata, min_cells=min_cells)

    # Normalizing total per cell
    if adata.X.max() < 50:
        print('Warning: adata_rna.X may have already been normalized, do not normalize, please check.')
        adata.layers['norm_data'] = adata.X.copy()
    else:
        sc.pp.normalize_total(adata, target_sum=target_sum)
        # Log1p transforming
        sc.pp.log1p(adata)
        adata.layers['norm_data'] = adata.X.copy()

    # Finding variable features for RNA adata
    if used_hvgs:
        if type(rna_n_top_features) == int:
            if rna_n_top_features > len(adata.var_names):
                rna_n_top_features = len(adata.var_names)
            if batch_key is not None:
                sc.pp.highly_variable_genes(adata, n_top_genes=rna_n_top_features, batch_key=batch_key)
            else:
                sc.pp.highly_variable_genes(adata, n_top_genes=rna_n_top_features)
                # sc.pp.highly_variable_genes(adata, min_mean=0.0125, max_mean=3, min_disp=0.5)
        elif type(rna_n_top_features) != int:
            if isinstance(rna_n_top_features, str):
                if os.path.isfile(rna_n_top_features):
                    rna_n_top_features = np.loadtxt(rna_n_top_features, dtype=str)
            adata = reindex(adata, rna_n_top_features)

    ## only HVGs
    use_highly_variable = "highly_variable" in adata.var
    adata_hvg = adata[:, adata.var["highly_variable"]].copy() if use_highly_variable else adata.copy()

    # scale data, clip values exceeding standard deviation 10.
    if adata_hvg.X.min() < 0:
        print('Warning: adata.X may have already been scaled, do not scale, please check.')
    else:
        sc.pp.scale(adata_hvg, max_value=10)
    # PCA
    sc.tl.pca(adata_hvg, svd_solver=svd_solver, n_comps=n_components)
    # 返回norm_data
    adata_hvg.X = adata_hvg.layers['norm_data'].copy() if "norm_data" in adata_hvg.layers.keys() else adata_hvg.X.copy()
    adata_hvg.obsm['feat'] = adata_hvg.obsm['X_pca'].copy()

    if used_pca_graph:
        # use scanpy functions to do the graph construction
        sc.pp.neighbors(adata_hvg, n_neighbors=n, metric=metric, use_rep='X_pca')

    return adata, adata_hvg

def preprocessing_atac(adata: AnnData,
                       used_hvgs: bool = True,
                       used_lsi_norm: bool = False,
                       used_lsi_graph: bool = True,
                       min_features: int = 100,
                       min_cells: int = 3,
                       atac_n_top_features=100000,  # or gene list
                       n: int = 6,
                       batch_key: str = 'batch',
                       metric: str = 'euclidean',
                       n_components: int = 50
                       ):
    """
    Preprocess scATAC data matrix.

    Parameters
    ----------
    adata :  AnnData object of shape `n_obs` × `n_vars`. Rows correspond to cells and columns to peaks/regions.
    filter_rate : float, optional
        Proportion for feature selection, by default 0.01
    """
    if min_features is None: min_features = 100
    if min_cells is None: min_cells = 3
    if atac_n_top_features is None: atac_n_top_features = 10000

    # Preprocessing
    if type(adata.X) != csr.csr_matrix:
        adata.X = scipy.sparse.csr_matrix(adata.X)

    sc.pp.filter_cells(adata, min_genes=min_features)
    sc.pp.filter_genes(adata, min_cells=min_cells)
    # adata.raw = adata

    ## TFIDF & LSI
    if used_lsi_norm:
        TFIDF_LSI(adata, n_comps=n_components, binarize=False, random_state=0)
        adata.layers['norm_data'] = adata.X.copy()
    else:
        # Normalizing total per cell
        if adata.X.max() < 2:
            print('Warning: adata_rna.X may have already been normalized, do not normalize, please check.')
            adata.layers['norm_data'] = adata.X.copy()
        else:
            sc.pp.normalize_per_cell(adata, counts_per_cell_after=1e4)
            # sc.pp.normalize_total(adata)
            # Log1p transforming
            sc.pp.log1p(adata)
            adata.layers['norm_data'] = adata.X.copy()

    # Finding variable features for ATAC adata
    if used_hvgs:
        if type(atac_n_top_features) == int:
            if atac_n_top_features > len(adata.var_names):
                atac_n_top_features = len(adata.var_names)
            if batch_key is not None:
                sc.pp.highly_variable_genes(adata, n_top_genes=atac_n_top_features, batch_key=batch_key)
            else:
                sc.pp.highly_variable_genes(adata, n_top_genes=atac_n_top_features)
                # sc.pp.highly_variable_genes(adata, min_mean=0.0125, max_mean=3, min_disp=0.5)
        elif type(atac_n_top_features) != int:
            if isinstance(atac_n_top_features, str):
                if os.path.isfile(atac_n_top_features):
                    atac_n_top_features = np.loadtxt(atac_n_top_features, dtype=str)
            adata = reindex(adata, atac_n_top_features)

    ## only HVGs
    use_highly_variable = "highly_variable" in adata.var
    adata_hvg = adata[:, adata.var["highly_variable"]].copy() if use_highly_variable else adata.copy()

    if used_lsi_graph:
        if not used_lsi_norm:
            # scale data, clip values exceeding standard deviation 10.
            if adata_hvg.X.min() < 0:
                print('Warning: adata.X may have already been scaled, do not scale, please check.')
            else:
                sc.pp.scale(adata_hvg, max_value=10)
            # PCA
            sc.tl.pca(adata_hvg, n_comps=n_components)
            # 返回norm_data
            adata_hvg.X = adata_hvg.layers['norm_data'].copy()
            adata_hvg.obsm['feat'] = adata_hvg.obsm['X_pca'].copy()

            # use scanpy functions to do the graph construction
            sc.pp.neighbors(adata_hvg, n_neighbors=n, metric=metric, use_rep='X_pca')
        else:
            # 返回norm_data
            adata_hvg.X = adata_hvg.layers['norm_data'].copy() if "norm_data" in adata_hvg.layers.keys() else adata_hvg.X.copy()
            adata_hvg.obsm['feat'] = adata_hvg.obsm['X_lsi'].copy()

            # use scanpy functions to do the graph construction
            sc.pp.neighbors(adata_hvg, n_neighbors=n, metric=metric, use_rep='X_lsi')

    return adata, adata_hvg

def preprocessing_adt(
        adata: AnnData,
        used_hvgs: bool = False,
        used_pca_graph: bool = True,
        rna_n_top_features=2000,  # or gene list
        n: int = 15,
        batch_key: str = 'batch',
        metric: str = 'euclidean',
        svd_solver: str = 'arpack',
        backed: bool = False
):
    """
    Preprocessing single-cell RNA-seq data

    Parameters
    ----------
    adata
        An AnnData matrice of shape n_obs x n_vars. Rows correspond to cells and columns to genes.
    min_features
        Filtered out cells that are detected in less than n genes. Default: 600.
    min_cells
        Filtered out genes that are detected in less than n cells. Default: 3.
    target_sum
        After normalization, each cell has a total count equal to target_sum. If None, total count of each cell equal to the median of total counts for cells before normalization.
    rna_n_top_features
        Number of highly-variable genes to keep. Default: 2000.

    Return
    -------
    The AnnData object after preprocessing.
    """
    # Preprocessing

    # if not issparse(adata.X):
    if type(adata.X) != csr.csr_matrix and (not backed) and (not adata.isbacked):
        adata.X = scipy.sparse.csr_matrix(adata.X)

    # Normalizing total per cell
    if adata.X.max() < 50:
        print('Warning: adata_adt.X may have already been normalized, do not normalize, please check.')
    else:
        adata = clr_normalize_each_cell(adata)
        adata.layers['norm_data'] = adata.X.copy()

    # Finding variable features for RNA adata
    if used_hvgs:
        if type(rna_n_top_features) == int:
            if rna_n_top_features > len(adata.var_names):
                rna_n_top_features = len(adata.var_names)
            if batch_key is not None:
                sc.pp.highly_variable_genes(adata, n_top_genes=rna_n_top_features, batch_key=batch_key)
            else:
                sc.pp.highly_variable_genes(adata, n_top_genes=rna_n_top_features)
                # sc.pp.highly_variable_genes(adata, min_mean=0.0125, max_mean=3, min_disp=0.5)
        elif type(rna_n_top_features) != int:
            if isinstance(rna_n_top_features, str):
                if os.path.isfile(rna_n_top_features):
                    rna_n_top_features = np.loadtxt(rna_n_top_features, dtype=str)
            adata = reindex(adata, rna_n_top_features)

    ## only HVGs
    use_highly_variable = "highly_variable" in adata.var
    adata_use = adata[:, adata.var["highly_variable"]].copy() if use_highly_variable else adata.copy()

    if used_pca_graph:
        # scale data, clip values exceeding standard deviation 10.
        if adata_use.X.min() < 0:
            print('Warning: adata.X may have already been scaled, do not scale, please check.')
        else:
            sc.pp.scale(adata_use, max_value=10)
        # PCA
        sc.tl.pca(adata_use, svd_solver=svd_solver)
        # use scanpy functions to do the graph construction
        sc.pp.neighbors(adata_use, n_neighbors=n, metric=metric, use_rep='X_pca')
        # 返回norm_data
        if adata.X.max() < 50:
            pass
        else:
            adata_use.X = adata_use.layers['norm_data'].copy() if "norm_data" in adata_use.layers.keys() else adata_use.X.copy()
        adata_use.obsm['feat'] = adata_use.obsm['X_pca'].copy()
    else:
        sc.pp.neighbors(adata_use, n_neighbors=n, metric=metric, use_rep='X')

    return adata, adata_use

### preprocessing main function
def preprocessing(
        adata: [AnnData, MuData],
        profile: str = 'RNA',
        data_type: str = 'Paired',
        sub_data_type: str = ['rna', 'atac'],
        batch_key: str = 'batch',
        weight = 0.8,
        used_hvgs: bool = True,
        graph_const_method: str = None,
        genome: str = None,
        use_gene_weigt: bool = True,
        use_top_pcs: bool = False,
        min_features: int = 600,
        min_cells: int = 3,
        target_sum: int = None,
        rna_n_top_features=None,  # or gene list
        atac_n_top_features=None,  # or gene list
        n_components: int = 50,
        svd_components_rna: int = 30,
        svd_components_atac: int = 30,
        cca_components: int = 20,
        cca_max_iter: int = 2000,
        randomized_svd: bool = False,
        filter_prop_initial: int = 0,
        filter_prop_refined: int = 0.3,
        filter_prop_propagated: int = 0,
        n_iters: int = 1,
        svd_runs: int = 1,
        n: int = 15,
        metric: str = 'euclidean',
        svd_solver: str = 'arpack',
        keep_mt: bool = False,
        backed: bool = False,
        verbose: bool = True
):
    """
    Preprocessing single-cell data

    Parameters
    ----------
    adata
        An AnnData matrice of shape n_obs x n_vars. Rows correspond to cells and columns to genes.
    profile
        Specify the single-cell profile type, RNA or ATAC, Default: RNA.
    min_features
        Filtered out cells that are detected in less than n genes. Default: 100.
    min_cells
        Filtered out genes that are detected in less than n cells. Default: 3.
    target_sum
        After normalization, each cell has a total count equal to target_sum. If None, total count of each cell equal to the median of total counts for cells before normalization.
    n_top_features
        Number of highly-variable genes to keep. Default: 2000.
    log
        If log, record each operation in the log file. Default: None.

    Return
    -------
    The AnnData object after preprocessing.

    """
    if profile == 'RNA':
        return preprocessing_rna(
            adata,
            min_features=min_features,
            min_cells=min_cells,
            target_sum=target_sum,
            used_hvgs=used_hvgs,
            used_pca_graph=True,
            rna_n_top_features=rna_n_top_features,
            n_components=n_components,
            n=n,
            batch_key=batch_key,
            metric=metric,
            svd_solver=svd_solver,
            keep_mt=keep_mt,
            backed=backed,
        )
    elif profile == 'ATAC':
        return preprocessing_atac(
            adata,
            min_features=min_features,
            min_cells=min_cells,
            used_hvgs=used_hvgs,
            used_lsi_graph=True,
            atac_n_top_features=atac_n_top_features,
            n_components=n_components,
            n=n,
            batch_key=batch_key,
            metric=metric
        )
    elif profile == 'ADT':
        return preprocessing_adt(
            adata,
            used_hvgs=used_hvgs,
            used_pca_graph=True,
            rna_n_top_features=rna_n_top_features,  # or gene list
            n=n,
            batch_key=batch_key,
            metric=metric,
            svd_solver=svd_solver,
            backed=backed
        )
    elif profile == 'multi-modal':
        # UnPaired TO DO
        assert data_type in ['Paired', 'UnPaired'], 'Data_type must be "Paired", or "UnPaired".'
        assert sub_data_type[0] in ['rna', 'atac', 'adt'] and sub_data_type[1] in ['rna', 'atac', 'adt'], \
            'Sub_data_type must be "rna", "atac", or "adt".'

        if len(sub_data_type) == 2:
            if sub_data_type[0] == 'rna' and sub_data_type[1] == 'atac':
                rna_adata = adata.mod['rna']
                atac_adata = adata.mod['atac']
                adt_adata = None
            elif sub_data_type[0] == 'rna' and sub_data_type[1] == 'adt':
                rna_adata = adata.mod['rna']
                atac_adata = None
                adt_adata = adata.mod['adt']
        else:
            ValueError('The length of sub_data_type must be 2, such as: ["rna", "atac"] or ["rna", "adt"].')

        if data_type == 'Paired':
            rna_adata, rna_adata_hvg = preprocessing_rna(
                rna_adata,
                min_features=min_features,
                min_cells=min_cells,
                target_sum=target_sum,
                used_hvgs=used_hvgs,
                used_pca_graph=True,
                rna_n_top_features=rna_n_top_features,
                n_components=n_components,
                n=n,
                batch_key=batch_key,
                metric=metric,
                svd_solver=svd_solver,
                keep_mt=keep_mt,
                backed=backed,
            )
            if atac_adata is not None:
                atac_adata, atac_adata_hvg = preprocessing_atac(
                    atac_adata,
                    min_features=min_features,
                    min_cells=min_cells,
                    used_hvgs=used_hvgs,
                    used_lsi_graph=True,
                    atac_n_top_features=atac_n_top_features,
                    n_components=n_components,
                    n=n,
                    batch_key=batch_key,
                    metric=metric
                )
                ## Concatenating different modalities
                adata_paired = ad.concat([rna_adata_hvg, atac_adata_hvg], axis=1)
                # 假设 obsm 的每个值都是形状相同的矩阵，可以通过 np.hstack 或其他方法进行合并
                rna_obsm = rna_adata_hvg.obsm
                atac_obsm = atac_adata_hvg.obsm
                # 只需要 feat key
                # combined_obsm = {key: np.hstack((rna_obsm[key], atac_obsm[key])) for key in rna_obsm.keys()}
                combined_obsm = np.hstack((rna_obsm['feat'], atac_obsm['feat']))
                # 将合并后的 obsm 信息添加到 adata_paired，但先确保索引匹配
                adata_paired.obsm['feat'] = pd.DataFrame(
                    combined_obsm,
                    index=adata_paired.obs.index  # 确保索引一致
                )

                ## the .obs layer is empty now, and we need to repopulate it
                rna_cols = rna_adata.obs.columns
                atac_cols = atac_adata.obs.columns

                rnaobs = rna_adata.obs.copy()
                rnaobs.columns = ["rna:" + x for x in rna_cols]
                atacobs = atac_adata.obs.copy()
                atacobs.columns = ["atac:" + x for x in atac_cols]
                adata_paired.obs = pd.merge(rnaobs, atacobs, left_index=True, right_index=True)

                ## 先将 scATAC 转换为基因活性矩阵
                # 定义保存文件的路径
                cache_path = "adata_ATAC_cache.h5ad"
                # 检查是否已经存在缓存文件
                if os.path.exists(cache_path):
                    # 如果缓存文件存在，直接加载
                    print("Gene activity matrix has been calculated, and loading cached adata_CG_atac object...")
                    adata_CG_atac = sc.read_h5ad(cache_path)
                else:
                    # 如果缓存文件不存在，执行耗时计算生成新的 adata
                    print('Convert peak to gene activity matrix, this might take a while...', flush=True)
                    print('`genome` parameter should be set correctly', flush=True)
                    print("Choose from {‘hg19’, ‘hg38’, ‘mm9’, ‘mm10’}", flush=True)
                    # 生成新的 adata_CG_atac 对象
                    adata_CG_atac = gene_scores(atac_adata, genome=genome, use_gene_weigt=use_gene_weigt, use_top_pcs=use_top_pcs)
                    # 将 adata_new 保存到缓存文件中
                    adata_CG_atac.write_h5ad(cache_path)
                    print(f"adata object cached at: {cache_path}")

                ## 交集
                common_genes = set(rna_adata.var_names).intersection(set(adata_CG_atac.var_names))
                print('There are {} common genes in RNA and ATAC datasets'.format(len(common_genes)))
                rna_adata_shared = rna_adata[:, list(common_genes)]
                atac_adata_shared = adata_CG_atac[:, list(common_genes)]

                # 通过cell matching 构建组学间的图结构
                print('To start performing cell matching for adjacency matrix of the graph, please wait...', flush=True)
                inter_connect = create_adj(rna_adata,
                                           atac_adata,
                                           rna_adata_shared,
                                           atac_adata_shared,
                                           data_type=data_type,
                                           rna_n_top_features=rna_n_top_features,
                                           atac_n_top_features=atac_n_top_features,
                                           batch_key=batch_key, svd_components1=svd_components_rna,
                                           svd_components2=svd_components_atac,
                                           cca_components=cca_components, cca_max_iter=cca_max_iter,
                                           randomized_svd=randomized_svd,
                                           filter_prop_initial=filter_prop_initial,
                                           filter_prop_refined=filter_prop_refined,
                                           filter_prop_propagated=filter_prop_propagated,
                                           n_iters=n_iters, svd_runs=svd_runs, verbose=verbose)

                # 设置权重
                # Iterate over keys in rna_obsp (assuming atac_obsp has the same keys)
                for key in rna_adata_hvg.obsp.keys():
                    # 计算加权平均的连通性矩阵
                    intra_connect = rna_adata_hvg.obsp[key] + atac_adata_hvg.obsp[key]

                adata_paired.obsp['connectivities'] = weight * inter_connect + (1 - weight) * intra_connect

                return adata_paired

            elif adt_adata is not None:
                adt_adata, adt_adata_hvg = preprocessing_adt(
                    adt_adata,
                    used_hvgs=used_hvgs,
                    used_pca_graph=True,
                    rna_n_top_features=rna_n_top_features,  # or gene list
                    n=n,
                    batch_key=batch_key,
                    metric=metric,
                    svd_solver=svd_solver,
                    backed=backed
                )
                ## Concatenating different modalities
                adata_paired = ad.concat([rna_adata_hvg, adt_adata_hvg], axis=1)

                ## the .obs layer is empty now, and we need to repopulate it
                rna_cols = rna_adata.obs.columns
                adt_cols = adt_adata.obs.columns

                rnaobs = rna_adata.obs.copy()
                rnaobs.columns = ["rna:" + x for x in rna_cols]
                adtobs = adt_adata.obs.copy()
                adtobs.columns = ["adt:" + x for x in adt_cols]
                adata_paired.obs = pd.merge(rnaobs, adtobs, left_index=True, right_index=True)

                ## 交集
                common_genes = set(rna_adata.var_names).intersection(set(adt_adata.var_names))
                print('There are {} common genes in RNA and ADT datasets'.format(len(common_genes)))
                rna_adata_shared = rna_adata[:, list(common_genes)]
                adt_adata_shared = adt_adata[:, list(common_genes)]

                ## 通过cell matching 构建组学间的图结构
                print('To start performing cell matching for adjacency matrix of the graph, please wait...', flush=True)
                inter_connect = create_adj(rna_adata,
                                           adt_adata,
                                           rna_adata_shared,
                                           adt_adata_shared,
                                           data_type=data_type,
                                           rna_n_top_features=rna_n_top_features,
                                           atac_n_top_features=atac_n_top_features,
                                           batch_key=batch_key, svd_components1=svd_components_rna,
                                           svd_components2=svd_components_atac,
                                           cca_components=cca_components, cca_max_iter=cca_max_iter,
                                           randomized_svd=randomized_svd,
                                           filter_prop_initial=filter_prop_initial,
                                           filter_prop_refined=filter_prop_refined,
                                           filter_prop_propagated=filter_prop_propagated,
                                           n_iters=n_iters, svd_runs=svd_runs, verbose=verbose)

                # Iterate over keys in rna_obsp (assuming adt_obsp has the same keys)
                for key in rna_adata_hvg.obsp.keys():
                    # 计算加权平均的连通性矩阵
                    intra_connect = rna_adata_hvg.obsp[key] + adt_adata_hvg.obsp[key]

                adata_paired.obsp['connectivities'] = weight * inter_connect + (1 - weight) * intra_connect

                return adata_paired

    elif profile == 'spatial':
        assert data_type in ['single-modal', 'multi-modal'], 'Data_type must be "single-modal", or "multi-modal".'

        if data_type == 'single-modal':
            _, rna_adata_hvg = preprocessing_rna(
                adata,
                min_features=min_features,
                min_cells=min_cells,
                target_sum=target_sum,
                used_hvgs=used_hvgs,
                used_pca_graph=True,
                rna_n_top_features=rna_n_top_features,
                n_components=n_components,
                n=n,
                batch_key=batch_key,
                metric=metric,
                svd_solver=svd_solver,
                keep_mt=True,
                backed=backed,
            )
            # Construct Spatial Graph
            if graph_const_method == 'mu_std':
                spatial_adj = graph_construction(adata, mode='mu_std', k=n, batch_key=batch_key)
                # spatial_adj = graph_dict #.toarray()
            elif graph_const_method == 'Radius':
                spatial_adj = graph_construction(adata, mode='Radius', k=n, batch_key=batch_key)
            elif graph_const_method == 'KNN':
                spatial_adj = graph_construction(adata, mode='KNN', k=n, batch_key=batch_key)
            elif graph_const_method == 'Squidpy':
                import squidpy as sq
                sq.gr.spatial_neighbors(adata, coord_type="generic", n_neighs=n)
                # Make adjacency matrix symmetric
                adata.obsp['spatial_connectivities'] = (
                    adata.obsp['spatial_connectivities'].maximum(
                        adata.obsp['spatial_connectivities'].T))
                spatial_adj = adata.obsp['spatial_connectivities'] #.toarray()

            # Ensure adj is a csr_matrix
            if not isinstance(spatial_adj, csr_matrix):
                spatial_adj = csr_matrix(spatial_adj)

            expr_adj = rna_adata_hvg.obsp['connectivities']
            # Ensure adj is a csr_matrix
            if not isinstance(expr_adj, csr_matrix):
                expr_adj = csr_matrix(expr_adj)

            # Validate adjacency matrix symmetry
            if (spatial_adj.getnnz() != spatial_adj.T.getnnz()):
                raise ImportError("The spatial_adj adjacency matrix has to be symmetric.")
            # Validate adjacency matrix symmetry
            if (expr_adj.getnnz() != expr_adj.T.getnnz()):
                raise ImportError("The expr_adj adjacency matrix has to be symmetric.")

            # 计算加权平均的连通性矩阵
            rna_adata_hvg.obsp['spatial_connectivities'] = spatial_adj
            rna_adata_hvg.obsp['connectivities'] = weight * spatial_adj + (1 - weight) * expr_adj

            return rna_adata_hvg

        elif data_type == 'multi-modal':
            assert sub_data_type[0] in ['rna', 'atac', 'adt'] and sub_data_type[1] in ['rna', 'atac', 'adt'], 'Sub_data_type must be "rna", "atac", or "adt".'

            if len(sub_data_type) == 2:
                if sub_data_type[0] == 'rna' and sub_data_type[1] == 'atac':
                    rna_adata = adata.mod['rna']
                    atac_adata = adata.mod['atac']
                    adt_adata = None
                elif sub_data_type[0] == 'rna' and sub_data_type[1] == 'adt':
                    rna_adata = adata.mod['rna']
                    atac_adata = None
                    adt_adata = adata.mod['adt']
            else:
                ValueError('The length of sub_data_type must be 2, such as: ["rna", "atac"] or ["rna", "adt"].')

            rna_adata, rna_adata_hvg = preprocessing_rna(
                rna_adata,
                min_features=min_features,
                min_cells=min_cells,
                target_sum=target_sum,
                used_hvgs=used_hvgs,
                used_pca_graph=True,
                rna_n_top_features=rna_n_top_features,
                n_components=n_components,
                n=n,
                batch_key=batch_key,
                metric=metric,
                svd_solver=svd_solver,
                keep_mt=True,
                backed=backed,
            )
            if atac_adata is not None:
                atac_adata, atac_adata_hvg = preprocessing_atac(
                    atac_adata,
                    min_features=min_features,
                    min_cells=min_cells,
                    used_hvgs=used_hvgs,
                    used_lsi_graph=True,
                    atac_n_top_features=atac_n_top_features,
                    n_components=n_components,
                    n=n,
                    batch_key=batch_key,
                    metric=metric
                )
                ## Concatenating different modalities
                adata_paired = ad.concat([rna_adata_hvg, atac_adata_hvg], axis=1)
                # 假设 obsm 的每个值都是形状相同的矩阵，可以通过 np.hstack 或其他方法进行合并
                rna_obsm = rna_adata_hvg.obsm
                atac_obsm = atac_adata_hvg.obsm
                # 只需要 feat key
                # combined_obsm = {key: np.hstack((rna_obsm[key], atac_obsm[key])) for key in rna_obsm.keys()}
                combined_obsm = np.hstack((rna_obsm['feat'], atac_obsm['feat']))
                # 将合并后的 obsm 信息添加到 adata_paired，但先确保索引匹配
                adata_paired.obsm['feat'] = pd.DataFrame(
                    combined_obsm,
                    index=adata_paired.obs.index  # 确保索引一致
                )
                adata_paired.obsm['spatial'] = rna_adata.obsm['spatial']

                ## the .obs layer is empty now, and we need to repopulate it
                rna_cols = rna_adata.obs.columns
                atac_cols = atac_adata.obs.columns

                rnaobs = rna_adata.obs.copy()
                rnaobs.columns = ["rna:" + x for x in rna_cols]
                atacobs = atac_adata.obs.copy()
                atacobs.columns = ["atac:" + x for x in atac_cols]
                adata_paired.obs = pd.merge(rnaobs, atacobs, left_index=True, right_index=True)

                ## 先将 scATAC 转换为基因活性矩阵
                if len(atac_adata.var_names) > 50000:
                    # 定义保存文件的路径
                    cache_path = "adata_ATAC_cache.h5ad"
                    # 检查是否已经存在缓存文件
                    if os.path.exists(cache_path):
                        # 如果缓存文件存在，直接加载
                        print("Gene activity matrix has been calculated, and loading cached adata_CG_atac object...")
                        adata_CG_atac = sc.read_h5ad(cache_path)
                    else:
                        # 如果缓存文件不存在，执行耗时计算生成新的 adata
                        print('Convert peak to gene activity matrix, this might take a while...', flush=True)
                        print('`genome` parameter should be set correctly', flush=True)
                        print("Choose from {‘hg19’, ‘hg38’, ‘mm9’, ‘mm10’}", flush=True)
                        # 生成新的 adata_CG_atac 对象
                        adata_CG_atac = gene_scores(atac_adata, genome=genome, use_gene_weigt=use_gene_weigt,
                                                    use_top_pcs=use_top_pcs)
                        # 将 adata_new 保存到缓存文件中
                        adata_CG_atac.write_h5ad(cache_path)
                        print(f"adata object cached at: {cache_path}")

                    ## 交集
                    common_genes = set(rna_adata.var_names).intersection(set(adata_CG_atac.var_names))
                    print('There are {} common genes in RNA and ATAC datasets'.format(len(common_genes)))
                    rna_adata_shared = rna_adata[:, list(common_genes)]
                    atac_adata_shared = adata_CG_atac[:, list(common_genes)]
                else:
                    ## 交集
                    common_genes = set(rna_adata.var_names).intersection(set(atac_adata.var_names))
                    print('There are {} common genes in RNA and ATAC datasets'.format(len(common_genes)))
                    rna_adata_shared = rna_adata[:, list(common_genes)]
                    atac_adata_shared = atac_adata[:, list(common_genes)]

                # 通过cell matching 构建组学间的图结构
                print('To start performing cell matching for adjacency matrix of the graph, please wait...', flush=True)
                inter_connect = create_adj(rna_adata,
                                           atac_adata,
                                           rna_adata_shared,
                                           atac_adata_shared,
                                           data_type=data_type,
                                           rna_n_top_features=rna_n_top_features,
                                           atac_n_top_features=atac_n_top_features,
                                           batch_key=batch_key, svd_components1=svd_components_rna,
                                           svd_components2=svd_components_atac,
                                           cca_components=cca_components, cca_max_iter=cca_max_iter,
                                           randomized_svd=randomized_svd,
                                           filter_prop_initial=filter_prop_initial,
                                           filter_prop_refined=filter_prop_refined,
                                           filter_prop_propagated=filter_prop_propagated,
                                           n_iters=n_iters, svd_runs=svd_runs, verbose=verbose)

                # Construct Spatial Graph
                if graph_const_method == 'mu_std':
                    spatial_adj = graph_construction(rna_adata_hvg, mode='mu_std', k=n, batch_key=batch_key)
                elif graph_const_method == 'Radius':
                    spatial_adj = graph_construction(adata, mode='Radius', k=n, batch_key=batch_key)
                elif graph_const_method == 'KNN':
                    spatial_adj = graph_construction(adata, mode='KNN', k=n, batch_key=batch_key)
                elif graph_const_method == 'Squidpy':
                    import squidpy as sq
                    sq.gr.spatial_neighbors(rna_adata_hvg, coord_type="generic", n_neighs=n)
                    spatial_adj = rna_adata_hvg.obsp['spatial_connectivities'] #.toarray()

                # Ensure adj is a csr_matrix
                if not isinstance(spatial_adj, csr_matrix):
                    spatial_adj = csr_matrix(spatial_adj)
                if not isinstance(inter_connect, csr_matrix):
                    inter_connect = csr_matrix(inter_connect)

                # Iterate over keys in rna_obsp (assuming adt_obsp has the same keys)
                # for key in rna_adata_hvg.obsp.keys():
                #     # 计算加权平均的连通性矩阵
                #     intra_connect = rna_adata_hvg.obsp[key] + atac_adata_hvg.obsp[key]

                # 计算加权平均的连通性矩阵
                adata_paired.obsp['spatial_connectivities'] = spatial_adj
                adata_paired.obsp['connectivities'] = weight * spatial_adj + (1 - weight) * inter_connect

                return adata_paired

            elif adt_adata is not None:
                adt_adata, adt_adata_hvg = preprocessing_adt(
                    adt_adata,
                    used_hvgs=used_hvgs,
                    used_pca_graph=True,
                    rna_n_top_features=rna_n_top_features,  # or gene list
                    n=n,
                    batch_key=batch_key,
                    metric=metric,
                    svd_solver=svd_solver,
                    backed=backed
                )
                ## Concatenating different modalities
                adata_paired = ad.concat([rna_adata_hvg, adt_adata_hvg], axis=1)
                # 假设 obsm 的每个值都是形状相同的矩阵，可以通过 np.hstack 或其他方法进行合并
                rna_obsm = rna_adata_hvg.obsm
                adt_obsm = adt_adata_hvg.obsm
                # 只需要 feat key
                # combined_obsm = {key: np.hstack((rna_obsm[key], atac_obsm[key])) for key in rna_obsm.keys()}
                combined_obsm = np.hstack((rna_obsm['feat'], adt_obsm['feat']))
                # 将合并后的 obsm 信息添加到 adata_paired，但先确保索引匹配
                adata_paired.obsm['feat'] = pd.DataFrame(
                    combined_obsm,
                    index=adata_paired.obs.index  # 确保索引一致
                )
                adata_paired.obsm['spatial'] = rna_adata.obsm['spatial']

                ## the .obs layer is empty now, and we need to repopulate it
                rna_cols = rna_adata.obs.columns
                adt_cols = adt_adata.obs.columns

                rnaobs = rna_adata.obs.copy()
                rnaobs.columns = ["rna:" + x for x in rna_cols]
                adtobs = adt_adata.obs.copy()
                adtobs.columns = ["adt:" + x for x in adt_cols]
                adata_paired.obs = pd.merge(rnaobs, adtobs, left_index=True, right_index=True)

                ## 交集
                common_genes = set(rna_adata.var_names).intersection(set(adt_adata.var_names))
                print('There are {} common genes in RNA and ADT datasets'.format(len(common_genes)))
                rna_adata_shared = rna_adata[:, list(common_genes)]
                adt_adata_shared = adt_adata[:, list(common_genes)]

                ## 通过cell matching 构建组学间的图结构
                print('To start performing cell matching for adjacency matrix of the graph, please wait...', flush=True)
                inter_connect = create_adj(rna_adata_hvg,
                                           adt_adata,
                                           rna_adata_shared,
                                           adt_adata_shared,
                                           data_type=data_type,
                                           rna_n_top_features=rna_n_top_features,
                                           atac_n_top_features=atac_n_top_features,
                                           batch_key=batch_key, svd_components1=svd_components_rna,
                                           svd_components2=svd_components_atac,
                                           cca_components=cca_components, cca_max_iter=cca_max_iter,
                                           randomized_svd=randomized_svd,
                                           filter_prop_initial=filter_prop_initial,
                                           filter_prop_refined=filter_prop_refined,
                                           filter_prop_propagated=filter_prop_propagated,
                                           n_iters=n_iters, svd_runs=svd_runs, verbose=verbose)

                # Construct Spatial Graph
                if graph_const_method == 'mu_std':
                    spatial_adj = graph_construction(rna_adata_hvg, mode='mu_std', k=n, batch_key=batch_key)
                elif graph_const_method == 'Radius':
                    spatial_adj = graph_construction(adata, mode='Radius', k=n, batch_key=batch_key)
                elif graph_const_method == 'KNN':
                    spatial_adj = graph_construction(adata, mode='KNN', k=n, batch_key=batch_key)
                elif graph_const_method == 'Squidpy':
                    import squidpy as sq
                    sq.gr.spatial_neighbors(rna_adata_hvg, coord_type="generic", n_neighs=n)
                    spatial_adj = rna_adata_hvg.obsp['spatial_connectivities'] #.toarray()

                # Ensure adj is a csr_matrix
                if not isinstance(spatial_adj, csr_matrix):
                    spatial_adj = csr_matrix(spatial_adj)
                if not isinstance(inter_connect, csr_matrix):
                    inter_connect = csr_matrix(inter_connect)

                # 计算加权平均的连通性矩阵
                adata_paired.obsp['spatial_connectivities'] = spatial_adj
                adata_paired.obsp['connectivities'] = weight * spatial_adj + (1 - weight) * inter_connect
                return adata_paired
    else:
        raise ValueError("Not support profile: `{}` yet".format(profile))


"""
Sphinx extension to add ReadTheDocs-style "Edit on GitHub" links to the
sidebar.
"""

import os
import warnings

__licence__ = "BSD (3 clause)"


# def get_github_repo(app, path):
#     if path.endswith(".ipynb"):
#         return app.config.github_nb_repo, "/"
#     return app.config.github_repo, "/docs/source/"


def html_page_context(app, pagename, templatename, context, doctree):
    if templatename != "page.html":
        return

    if doctree is not None:
        path = os.path.relpath(doctree.get("source"), app.builder.srcdir)
        if path.endswith(".ipynb"):
            context["display_github"] = True
            context["github_user"] = "zhou-1314"
            context["github_repo"] = "Garfield_tutorials"
            context["github_version"] = "main"
            if path.endswith("rna_10x_mouse_brain_1p3M.ipynb"):
                context["conf_py_path"] = "/v1.1/"
            else:
                context["conf_py_path"] = "/v1.0/"
        else:
            context["display_github"] = True
            context["github_user"] = "zhou-1314"
            context["github_repo"] = "Garfield"
            context["github_version"] = "main"
            context["conf_py_path"] = "/docs/source/"

def setup(app):
    app.add_config_value("github_nb_repo", "", True)
    app.add_config_value("github_repo", "", True)
    app.connect("html-page-context", html_page_context)



"""
Utility functions for matching
"""
import pandas as pd
import numpy as np
import pynndescent
from scipy.optimize import linear_sum_assignment

from . import _utils as utils

# modify from Maxfuse: https://github.com/shuxiaoc/maxfuse/blob/main/maxfuse/match_utils.py#L273
def match_cells(arr1, arr2, base_dist=None, wt_on_base_dist=0, verbose=True):
    """
    Get matching between arr1 and arr2 using linear assignment, the distance is 1 - Pearson correlation.

    Parameters
    ----------
    arr1: np.array of shape (n_samples1, n_features)
        The first data matrix
    arr2: np.array of shape (n_samples2, n_features)
        The second data matrix
    base_dist: None or np.ndarray of shape (n_samples1, n_samples2)
        Baseline distance matrix
    wt_on_base_dist: float between 0 and 1
        The final distance matrix to use is (1-wt_on_base_dist) * dist[arr1, arr2] + wt_on_base_dist * base_dist
    verbose: bool, default=True
        Whether to print the progress

    Returns
    -------
    rows, cols, vals: list
        Each matched pair of rows[i], cols[i], their distance is vals[i]
    """
    if verbose:
        print('Start the matching process...', flush=True)
        print('Computing the distance matrix...', flush=True)
    dist = utils.cdist_correlation(arr1, arr2)
    if base_dist is not None:
        if verbose:
            print(
                f'Interpolating {1-wt_on_base_dist} * dist[arr1, arr2] + {wt_on_base_dist} * base_dist...',
                flush=True
            )
        dist = (1-wt_on_base_dist) * dist + wt_on_base_dist * base_dist
    if verbose:
        print('Solving linear assignment...', flush=True)
    rows, cols = linear_sum_assignment(dist)
    if verbose:
        print('Linear assignment completed!', flush=True)

    return rows, cols, np.array([dist[i, j] for i, j in zip(rows, cols)])


def get_initial_matching(
        arr1, arr2,
        randomized_svd=True,
        svd_runs=1,
        svd_components1=30, svd_components2=30,
        verbose=True
):
    """
    Assume the features of arr1 and arr2 are column-wise directly comparable,
    obtain a matching by minimizing the correlation distance between arr1 and arr2.

    Parameters
    ----------
    arr1: np.array of shape (n_samples1, n_features1)
        The first data matrix.
    arr2: np.array of shape (n_samples2, n_features2)
        The second data matrix.
    randomized_svd: bool, default=False
        Whether to use randomized svd.
    svd_runs: int, default=1
        Randomized SVD will result in different runs,
        so if randomized_svd=True, perform svd_runs many randomized SVDs, and pick the one with the
        smallest Frobenious reconstruction error.
        If randomized_svd=False, svd_runs is forced to be 1.
    svd_components1: None or int
        If None, then do not do SVD,
        else, number of components to keep when doing SVD de-noising for the first data matrix.
    svd_components2: None or int
        Same as svd_components1 but for the second data matrix.
    verbose: bool, default=True
        Whether to print the progress.

    Returns
    -------
    matching: list of length 3
        rows, cols, vals = matching,
        Each matched pair is rows[i], cols[i], their distance is vals[i].
    """
    # assert arr1.shape[1] == arr2.shape[1]

    arr1 = utils.convert_to_numpy(arr1)
    arr2 = utils.convert_to_numpy(arr2)
    arr1, arr2 = utils.drop_zero_variability_columns(arr_lst=[arr1, arr2])

    # 确保 SVD 分量数不超过特征数，如果超过则调整并发出警告
    if arr1.shape[1] < svd_components1 or arr2.shape[1] < svd_components2:
        svd_components1, svd_components2 = arr1.shape[1]-1, arr2.shape[1]-1
        # 发出警告信息
        print(
            f"Warning: svd_components1 and svd_components2 were set too large and have been adjusted to {svd_components1} and {svd_components2}, "
            "which are the number of features in arr1 and arr2 respectively."
        )

    # denoising
    if verbose:
        print("Denoising the data...", flush=True)
    arr1_svd = utils.svd_denoise(
        arr=arr1, n_components=svd_components1, randomized=randomized_svd,
        n_runs=svd_runs
    )
    arr2_svd = utils.svd_denoise(
        arr=arr2, n_components=svd_components2, randomized=randomized_svd,
        n_runs=svd_runs
    )

    # if verbose:
    #     print('Normalizing and reducing the dimension of the data...', flush=True)
    # arr1_svd = utils.svd_embedding(
    #     arr=arr1, n_components=svd_components1,
    #     randomized=randomized_svd, n_runs=svd_runs
    # )
    # arr2_svd = utils.svd_embedding(
    #     arr=arr2, n_components=svd_components2,
    #     randomized=randomized_svd, n_runs=svd_runs
    # )

    res = match_cells(arr1=arr1_svd, arr2=arr2_svd, verbose=verbose)
    if verbose:
        print('Initial matching completed!', flush=True)

    return res

def get_refined_matching_one_iter(
        init_matching, arr1, arr2,
        filter_prop=0,
        cca_components=15,
        cca_max_iter=2000,
        verbose=True
):
    """
    Run one iteration of CCA refinement.

    Parameters
    ----------
    init_matching: list
        init_matching[0][i], init_matching[1][i] is a matched pair,
        and init_matching[2][i] is the distance for this pair
    arr1: np.array of shape (n_samples1, n_features1)
        The first data matrix.
    arr2: np.array of shape (n_samples2, n_features2)
        The second data matrix.
    filter_prop: float, default=0
        Proportion of matched pairs to discard before feeding into refinement iterations.
    cca_components: int, default=15
        Number of CCA components.
    cca_max_iter: int, default=2000
        Maximum number of CCA iterations.
    verbose: bool, default=True
        Whether to print the

    Returns
    -------
    rows, cols, vals: list
        Each matched pair of rows[i], cols[i], their distance is vals[i]
    """
    if verbose:
        print('Fitting CCA...', flush=True)
    arr1_cca, arr2_cca, _ = utils.cca_embedding(
        arr1=arr1, arr2=arr2,
        init_matching=init_matching, filter_prop=filter_prop, n_components=cca_components, max_iter=cca_max_iter
    )

    return match_cells(
        arr1=arr1_cca, arr2=arr2_cca, verbose=verbose
    )


def get_refined_matching(
        init_matching, arr1, arr2,
        randomized_svd=False, svd_runs=1,
        svd_components1=None, svd_components2=None,
        n_iters=3, filter_prop=0,
        cca_components=20,
        cca_max_iter=2000,
        verbose=True
):
    """
    Refinement of init_matching.

    Parameters
    ----------
    init_matching: list
        init_matching[0][i], init_matching[1][i] is a matched pair,
        and init_matching[2][i] is the distance for this pair.
    arr1: np.array of shape (n_samples1, n_features1)
        The first data matrix.
    arr2: np.array of shape (n_samples2, n_features2)
        The second data matrix.
    randomized_svd: bool, default=False
        Whether to use randomized SVD
    svd_runs: int, default=1
        Randomized SVD will result in different runs,
        so if randomized_svd=True, perform svd_runs many randomized SVDs, and pick the one with the
        smallest Frobenious reconstruction error.
        If randomized_svd=False, svd_runs is forced to be 1.
    svd_components1: None or int
        If None, then do not do SVD,
        else, number of components to keep when doing SVD de-noising for the first data matrix
        before feeding into CCA.
    svd_components2: None or int
        Same as svd_components1 but for the second data matrix.
    n_iters: int, default=3
        Number of refinement iterations.
    filter_prop: float, default=0
        Proportion of matched pairs to discard before feeding into refinement iterations.
    cca_components: int, default=15
        Number of CCA components.
    cca_max_iter: int, default=2000,
        Maximum number of CCA iterations.
    verbose: bool, default=True
        Whether to print the progress.

    Returns
    -------
    matching: list of length 3
        rows, cols, vals = matching,
        Each matched pair is rows[i], cols[i], their distance is vals[i].
    """
    ns = [len(x) for x in init_matching]
    assert ns[0] == ns[1] == ns[2]
    assert isinstance(n_iters, int) and n_iters >= 1
    assert 0 <= int(ns[0] * filter_prop) < ns[0]

    assert 1 <= cca_components <= min(arr1.shape[1], arr2.shape[1])

    # 确保 SVD 分量数不超过特征数，如果超过则调整并发出警告
    if arr1.shape[1] < cca_components or arr2.shape[1] < cca_components:
        cca_components = min(arr1.shape[1], arr2.shape[1])
        # 发出警告信息
        print(
            f"Warning: cca_components were set too large and have been adjusted to {cca_components}, "
            "which are the min number of features of arr1 and arr2."
        )

    if verbose:
        print('Normalizing and reducing the dimension of the data...', flush=True)
    arr1 = utils.svd_embedding(
        arr=arr1, n_components=svd_components1,
        randomized=randomized_svd, n_runs=svd_runs
    )
    arr2 = utils.svd_embedding(
        arr=arr2, n_components=svd_components2,
        randomized=randomized_svd, n_runs=svd_runs
    )

    # assert arr1.shape[1] == arr2.shape[1]
    # denoising
    # if verbose:
    #     print("Denoising the data...", flush=True)
    # arr1 = utils.svd_denoise(
    #     arr=arr1, n_components=svd_components1, randomized=randomized_svd,
    #     n_runs=svd_runs
    # )
    # arr2 = utils.svd_denoise(
    #     arr=arr2, n_components=svd_components2, randomized=randomized_svd,
    #     n_runs=svd_runs
    # )

    # prepare the distance matrix used in the initial matching
    cca_matching = init_matching
    # iterative refinement
    for it in range(n_iters):
        if verbose:
            print('Now at iteration {}:'.format(it), flush=True)
        cca_matching = get_refined_matching_one_iter(
            init_matching=cca_matching,
            arr1=arr1, arr2=arr2,
            filter_prop=filter_prop,
            cca_components=cca_components,
            cca_max_iter=cca_max_iter, verbose=verbose
        )

    if verbose:
        print('Refined matching completed!', flush=True)
    return cca_matching

def fit_svd_on_full_data(arr1, arr2, randomized_svd=False, svd_runs=1,
                         svd_components1=None, svd_components2=None):
    """Perform SVD on full self.active_arr1 and self.active_arr2 and save the functions that reduce the dimension
    of the data.
    """
    if svd_components1 is not None:
        u1, s1, vh1 = utils.robust_svd(
            arr=arr1, n_components=svd_components1,
            randomized=randomized_svd, n_runs=svd_runs
        )
        rotation_before_cca_on_active_arr1 = lambda arr: arr @ vh1.T
    else:
        rotation_before_cca_on_active_arr1 = lambda arr: arr

    if svd_components2 is not None:
        u2, s2, vh2 = utils.robust_svd(
            arr=arr2, n_components=svd_components2,
            randomized=randomized_svd, n_runs=svd_runs
        )
        rotation_before_cca_on_active_arr2 = lambda arr: arr @ vh2.T
    else:
        rotation_before_cca_on_active_arr2 = lambda arr: arr

    return rotation_before_cca_on_active_arr1, rotation_before_cca_on_active_arr2


def get_nearest_neighbors(query_arr, target_arr, svd_components=None, randomized_svd=False, svd_runs=1,
                          metric='correlation'):
    """
    For each row in query_arr, compute its nearest neighbor in target_arr.

    Parameters
    ----------
    query_arr: np.array of shape (n_samples1, n_features)
        The query data matrix.
    target_arr: np.array of shape (n_samples2, n_features)
        The target data matrix.
    svd_components: None or int, default=None
        If not None, will first conduct SVD to reduce the dimension
        of the vertically stacked version of query_arr and target_arr.
    randomized_svd: bool, default=False
        Whether to use randomized SVD.
    svd_runs: int, default=1
        Run multiple instances of SVD and select the one with the lowest Frobenious reconstruction error.
    metric: string, default='correlation'
        The metric to use in nearest neighbor search.

    Returns
    -------
    neighbors: np.array of shape (n_samples1)
        The i-th element is the index in target_arr to whom the i-th row of query_arr is closest to.
    dists: np.array of shape (n_samples1)
        The i-th element is the distance corresponding to neighbors[i].
    """
    arr = np.vstack([query_arr, target_arr])
    arr = utils.svd_embedding(
        arr=arr, n_components=svd_components,
        randomized=randomized_svd,
        n_runs=svd_runs
    )
    query_arr = arr[:query_arr.shape[0], :]
    pivot_arr = arr[query_arr.shape[0]:, :]
    # approximate nearest neighbor search
    index = pynndescent.NNDescent(pivot_arr, n_neighbors=100, metric=metric)
    neighbors, dists = index.query(query_arr, k=50)
    neighbors, dists = neighbors[:, 0], dists[:, 0]
    return neighbors, dists


def address_matching_redundancy(matching, direction='both'):
    """
    Make a potentially multiple-to-multiple matching to an one-to-one matching according to order.

    Parameters
    ----------
    matching: list of length three
        rows, cols, vals = matching: list
        Each matched pair of rows[i], cols[i], their score (the larger, the better) is vals[i]
    direction: 'left' or 'right' or 'both', default=both
        If None, do nothing;
        If (1, 2), then the redundancy is addressed by making matching
        an injective map from the first dataset to the second;
        if (2, 1), do the other way around.

    Returns
    -------
    rows, cols, vals: list
        Each matched pair of rows[i], cols[i], their score is vals[i].
    """
    if direction == 'both':
        return matching
    res = [[], [], []]
    if direction == 'left':
        idx1_to_idx2 = dict()
        idx1_to_score = dict()
        for i, j, score in zip(matching[0], matching[1], matching[2]):
            if i not in idx1_to_idx2:
                idx1_to_idx2[i] = j
                idx1_to_score[i] = score
            elif score > idx1_to_score[i]:
                idx1_to_idx2[i] = j
                idx1_to_score[i] = score
        for idx1, idx2 in idx1_to_idx2.items():
            res[0].append(idx1)
            res[1].append(idx2)
            res[2].append(idx1_to_score[idx1])
    elif direction == 'right':
        idx2_to_idx1 = dict()
        idx2_to_score = dict()
        for i, j, score in zip(matching[0], matching[1], matching[2]):
            if j not in idx2_to_idx1:
                idx2_to_idx1[j] = i
                idx2_to_score[j] = score
            elif score > idx2_to_score[j]:
                idx2_to_idx1[j] = i
                idx2_to_score[j] = score
        for idx2, idx1 in idx2_to_idx1.items():
            res[0].append(idx1)
            res[1].append(idx2)
            res[2].append(idx2_to_score[idx2])
    else:
        raise NotImplementedError('direction must be in {both, left, right}.')

    return res



from .loss import (compute_omics_recon_mse_loss,
                     compute_edge_recon_loss,
                     compute_kl_reg_loss,
                     compute_contrastive_instanceloss,
                     compute_contrastive_clusterloss,
                     compute_omics_recon_mmd_loss
                   )
from .GNNModelVAE import GNNModelVAE

__all__ = ["compute_omics_recon_mse_loss",
           "compute_edge_recon_loss",
           "compute_kl_reg_loss",
           "compute_contrastive_instanceloss",
           "compute_contrastive_clusterloss",
           "GNNModelVAE"
           ]


"""
This module contains helper functions for the ´models´ subpackage.
"""

import logging
import os
import pickle
import dill
from collections import OrderedDict
from typing import Optional, Tuple, Literal

import numpy as np
import pandas as pd
import torch
import scanpy as sc
import anndata
import anndata as ad
from anndata import AnnData, concat
from scipy.sparse import csr_matrix, hstack
from sklearn.neighbors import KNeighborsTransformer

logger = logging.getLogger(__name__)


def save_model_with_fallback(model, file_path):
    try:
        # 尝试使用 pickle 保存
        with open(file_path, "wb") as f:
            pickle.dump(model, f)
        print(f"Model saved successfully using pickle at {file_path}")
    except (AttributeError, pickle.PicklingError) as e:
        # 如果 pickle 保存失败，捕获异常并使用 joblib
        print(f"Pickle failed with error: {e}, switching to dill...")
        with open(file_path, "wb") as f:
            dill.dump(model, f)
        print(f"Model saved successfully using dill at {file_path}")

def load_model_with_fallback(file_path):
    try:
        # 尝试使用 pickle 加载
        with open(file_path, "rb") as f:
            model = pickle.load(f)
        print(f"Model loaded successfully using pickle from {file_path}")
    except (AttributeError, pickle.UnpicklingError, EOFError) as e:
        # 如果 pickle 加载失败，捕获异常并使用 joblib
        print(f"Pickle failed with error: {e}, switching to dill...")
        with open(file_path, "rb") as f:
            model = dill.load(f)
        print(f"Model loaded successfully using dill from {file_path}")

    return model


def load_saved_files(dir_path: str,
                     query_adata: Optional[AnnData] = None,
                     ref_adata_name: str = "adata_ref.h5ad",
                     map_location: Optional[Literal["cpu", "cuda"]] = None
                     ) -> Tuple[OrderedDict, dict, np.ndarray, ad.AnnData]:
    """
    Helper to load saved model files.

    Parts of the implementation are adapted from
    https://github.com/scverse/scvi-tools/blob/master/scvi/model/base/_utils.py#L55
    (01.10.2022)

    Parameters
    ----------
    dir_path:
        Path where the saved model files are stored.
    load_adata:
        If `True`, also load the stored AnnData object.
    adata_file_name:
        File name under which the AnnData object is saved.
    load_adata_atac:
        If `True`, also load the stored ATAC AnnData object.
    adata_atac_file_name:
        File name under which the ATAC AnnData object is saved.
    map_location:
        Memory location where to map the model files to.

    Returns
    ----------
    model_state_dict:
        The stored model state dict.
    var_names:
        The stored variable names.
    attr_dict:
        The stored attributes.
    adata:
        The stored AnnData object.
    adata_atac:
        The stored ATAC AnnData object.
    """
    attr_path = os.path.join(dir_path, "attr.pkl")
    adata_path = os.path.join(dir_path, ref_adata_name)
    var_names_path = os.path.join(dir_path, "var_names.csv")
    model_path = os.path.join(dir_path, "model_params.pt")

    if os.path.exists(adata_path):
        adata_ref = ad.read(adata_path)
        adata_ref.X = adata_ref.layers["counts"].copy() if "counts" in adata_ref.layers.keys() else adata_ref.X
    else:
        raise ValueError("Dir path contains no saved reference anndata")

    var_names = np.genfromtxt(var_names_path, delimiter=",", dtype=str)
    if query_adata is not None:
        query_adata = validate_var_names(query_adata, var_names)
        adata_concat = concat(
            [adata_ref, query_adata],
            label='projection',
            keys=['reference', 'query'],
            index_unique=None,
            join='outer'
        )
    else:
        adata_concat = None

    model_state_dict = torch.load(model_path, map_location=map_location)
    attr_dict = load_model_with_fallback(attr_path)
    # with open(attr_path, "rb") as handle:
    #     attr_dict = pickle.load(handle)
    return model_state_dict, var_names, attr_dict, adata_concat


def validate_var_names(adata, source_var_names):
    # Warning for gene percentage
    user_var_names = adata.var_names
    try:
        percentage = (len(user_var_names.intersection(source_var_names)) / len(user_var_names)) * 100
        percentage = round(percentage, 4)
        if percentage != 100:
            logger.warning(f"WARNING: Query shares {percentage}% of its genes with the reference."
                           "This may lead to inaccuracy in the results.")
    except Exception:
        logger.warning("WARNING: Something is wrong with the reference genes.")

    user_var_names = user_var_names.astype(str)
    new_adata = adata

    # Get genes in reference that are not in query
    ref_genes_not_in_query = []
    for name in source_var_names:
        if name not in user_var_names:
            ref_genes_not_in_query.append(name)

    if len(ref_genes_not_in_query) > 0:
        print("Query data is missing expression data of ",
              len(ref_genes_not_in_query),
              " genes which were contained in the reference dataset.")
        print("The missing information will be filled with zeroes.")

        filling_X = np.zeros((len(adata), len(ref_genes_not_in_query)))
        if isinstance(adata.X, csr_matrix):
            filling_X = csr_matrix(filling_X)  # support csr sparse matrix
            new_target_X = hstack((adata.X, filling_X))
        else:
            new_target_X = np.concatenate((adata.X, filling_X), axis=1)
        new_target_vars = adata.var_names.tolist() + ref_genes_not_in_query
        new_adata = AnnData(new_target_X, dtype="float32")
        new_adata.var_names = new_target_vars
        new_adata.obs = adata.obs.copy()

    if len(user_var_names) - (len(source_var_names) - len(ref_genes_not_in_query)) > 0:
        print(
            "Query data contains expression data of ",
            len(user_var_names) - (len(source_var_names) - len(ref_genes_not_in_query)),
            " genes that were not contained in the reference dataset. This information "
            "will be removed from the query data object for further processing.")

        # remove unseen gene information and order anndata
        new_adata = new_adata[:, source_var_names].copy()

    print(new_adata)
    return new_adata

def weighted_knn_trainer(train_adata, train_adata_emb, n_neighbors=50):
    """Trains a weighted KNN classifier on ``train_adata``.
    Parameters
    ----------
    train_adata: :class:`~anndata.AnnData`
        Annotated dataset to be used to train KNN classifier with ``label_key`` as the target variable.
    train_adata_emb: str
        Name of the obsm layer to be used for calculation of neighbors. If set to "X", anndata.X will be
        used
    n_neighbors: int
        Number of nearest neighbors in KNN classifier.
    """
    print(
        f"Weighted KNN with n_neighbors = {n_neighbors} ... "
    )
    k_neighbors_transformer = KNeighborsTransformer(
        n_neighbors=n_neighbors,
        mode="distance",
        algorithm="brute",
        metric="euclidean",
        n_jobs=-1,
    )
    if train_adata_emb == "X":
        train_emb = train_adata.X
    elif train_adata_emb in train_adata.obsm.keys():
        train_emb = train_adata.obsm[train_adata_emb]
    else:
        raise ValueError(
            "train_adata_emb should be set to either 'X' or the name of the obsm layer to be used!"
        )
    k_neighbors_transformer.fit(train_emb)
    return k_neighbors_transformer

def weighted_knn_transfer(
    query_adata,
    query_adata_emb,
    ref_adata_obs,
    label_keys,
    knn_model,
    threshold=1,
    pred_unknown=False,
    mode="package",
):
    """Annotates ``query_adata`` cells with an input trained weighted KNN classifier.
    Parameters
    ----------
    query_adata: :class:`~anndata.AnnData`
        Annotated dataset to be used to queryate KNN classifier. Embedding to be used
    query_adata_emb: str
        Name of the obsm layer to be used for label transfer. If set to "X",
        query_adata.X will be used
    ref_adata_obs: :class:`pd.DataFrame`
        obs of ref Anndata
    label_keys: str
        Names of the columns to be used as target variables (e.g. cell_type) in ``query_adata``.
    knn_model: :class:`~sklearn.neighbors._graph.KNeighborsTransformer`
        knn model trained on reference adata with weighted_knn_trainer function
    threshold: float
        Threshold of uncertainty used to annotating cells as "Unknown". cells with
        uncertainties higher than this value will be annotated as "Unknown".
        Set to 1 to keep all predictions. This enables one to later on play
        with thresholds.
    pred_unknown: bool
        ``False`` by default. Whether to annotate any cell as "unknown" or not.
        If `False`, ``threshold`` will not be used and each cell will be annotated
        with the label which is the most common in its ``n_neighbors`` nearest cells.
    mode: str
        Has to be one of "paper" or "package". If mode is set to "package",
        uncertainties will be 1 - P(pred_label), otherwise it will be 1 - P(true_label).
    """
    if not type(knn_model) == KNeighborsTransformer:
        raise ValueError(
            "knn_model should be of type sklearn.neighbors._graph.KNeighborsTransformer!"
        )

    if query_adata_emb == "X":
        query_emb = query_adata.X
    elif query_adata_emb in query_adata.obsm.keys():
        query_emb = query_adata.obsm[query_adata_emb]
    else:
        raise ValueError(
            "query_adata_emb should be set to either 'X' or the name of the obsm layer to be used!"
        )
    top_k_distances, top_k_indices = knn_model.kneighbors(X=query_emb)

    stds = np.std(top_k_distances, axis=1)
    stds = (2.0 / stds) ** 2
    stds = stds.reshape(-1, 1)

    top_k_distances_tilda = np.exp(-np.true_divide(top_k_distances, stds))

    weights = top_k_distances_tilda / np.sum(
        top_k_distances_tilda, axis=1, keepdims=True
    )
    cols = ref_adata_obs.columns[ref_adata_obs.columns.str.startswith(label_keys)]

    uncertainties = pd.DataFrame(columns=cols, index=query_adata.obs_names)
    pred_labels = pd.DataFrame(columns=cols, index=query_adata.obs_names)
    for i in range(len(weights)):
        for j in cols:
            y_train_labels = ref_adata_obs[j].values
            unique_labels = np.unique(y_train_labels[top_k_indices[i]])
            best_label, best_prob = None, 0.0
            for candidate_label in unique_labels:
                candidate_prob = weights[
                    i, y_train_labels[top_k_indices[i]] == candidate_label
                ].sum()
                if best_prob < candidate_prob:
                    best_prob = candidate_prob
                    best_label = candidate_label

            if pred_unknown:
                if best_prob >= threshold:
                    pred_label = best_label
                else:
                    pred_label = "Unknown"
            else:
                pred_label = best_label

            if mode == "package":
                uncertainties.iloc[i][j] = (max(1 - best_prob, 0))

            else:
                raise Exception("Inquery Mode!")

            pred_labels.iloc[i][j] = (pred_label)

    print("Label transfer finished!", flush=True)

    return pred_labels, uncertainties


"""
This module contains the GarphAnnTorchDataset class to provide a standardized
dataset format for the training of Garfield model.
"""

from typing import Literal, List, Optional, Tuple, Union

import numpy as np
import torch

import scipy.sparse as sp
from scipy.sparse import csr_matrix
from torch_sparse import SparseTensor

from anndata import AnnData
from mudata import MuData
from sklearn.preprocessing import LabelEncoder
from torch_geometric.utils import add_self_loops, remove_self_loops


class GraphAnnTorchDataset():
    """
    Spatially annotated torch dataset class to extract node features, node
    labels, adjacency matrix and edge indices in a standardized format from an
    AnnData object.

    Parameters
    ----------
    adata:
        AnnData object with counts stored in ´adata.layers[counts_key]´ or
        ´adata.X´ depending on ´counts_key´, and sparse adjacency matrix stored
        in ´adata.obsp[adj_key]´.
    adj_key:
        Key under which the sparse adjacency matrix is stored in ´adata.obsp´.
    self_loops:
        If ´True´, add self loops to the adjacency matrix to model autocrine
        communication.

    Returns
    ----------
    None
    """
    def __init__(self,
                 adata, # : AnnData,
                 label_name: str = None,
                 adj_key: Literal["spatial_connectivities", "connectivities"]="spatial_connectivities",
                 edge_label_adj_key: str = "edge_label_spatial_connectivities",
                 used_feat: bool = False,
                 self_loops: bool = True):
        super(GraphAnnTorchDataset, self).__init__()

        if not used_feat:
            x = adata.X
        else:
            x = np.array(adata.obsm['feat'])

        # Store features in dense format
        if sp.issparse(x):
            self.x = torch.tensor(x.toarray())
        else:
            self.x = torch.tensor(x)

        ## celltype label_encoder Todo 如果标签信息本身就是数值则不需要转换
        if label_name is not None:
            label_encoder = LabelEncoder()
            meta = np.array(adata.obs[label_name].astype('str'))
            meta = label_encoder.fit_transform(meta)
            # meta = meta.astype(np.float32)
            inverse = label_encoder.inverse_transform(range(0, np.max(meta) + 1))
            self.y = torch.from_numpy(meta.astype(np.int64))
        else:
            self.y = torch.from_numpy(np.array([1] * adata.shape[0]))

        # Store adjacency matrix in torch_sparse SparseTensor format
        if sp.issparse(adata.obsp[adj_key]):
            self.adj = sparse_mx_to_sparse_tensor(adata.obsp[adj_key])
        else:
            self.adj = sparse_mx_to_sparse_tensor(
                sp.csr_matrix(adata.obsp[adj_key]))

        # Store edge label adjacency matrix
        if edge_label_adj_key in adata.obsp:
            self.edge_label_adj = sp.csr_matrix(adata.obsp[edge_label_adj_key])
        else:
            self.edge_label_adj = None

        # Validate adjacency matrix symmetry
        if (self.adj.nnz() != self.adj.t().nnz()):
            raise ImportError("The input adjacency matrix has to be symmetric.")

        # 获取 edge_index
        self.edge_index = self.adj.to_torch_sparse_coo_tensor()._indices()
        # # 获取 edge_index 对应的value 或 weight
        # self.edge_weight = self.adj.storage.value() # 对应的value
        self.edge_weight = self.adj.to_torch_sparse_coo_tensor()._values()  # 非零索引对应的值
        # 确保 edge_weight 的形状是 (num_edges, 1)，因为 edge_attr 通常是 (num_edges, attr_dim)
        self.edge_weight = torch.unsqueeze(self.edge_weight, dim=-1)

        if self_loops:
            # Add self loops to account for autocrine communication
            # Remove self loops in case there are already before adding new ones
            self.edge_index, _ = remove_self_loops(self.edge_index)
            self.edge_index, _ = add_self_loops(self.edge_index,
                                                num_nodes=self.x.size(0))

            # 确保添加的自环数目正确
            new_edge_index_shape = self.edge_index.shape[1]
            original_edge_index_shape = self.edge_weight.shape[0]
            num_self_loops = new_edge_index_shape - original_edge_index_shape
            # 创建自环边的权重，假设权重值为1
            self_loops_weight = torch.ones((num_self_loops, 1), dtype=self.edge_weight.dtype)

            # 获取新增自环的边的数量
            # num_self_loops = self.x.size(0)
            # # 创建自环边的权重，假设权重值为1
            # self_loops_weight = torch.ones((num_self_loops, 1), dtype=self.edge_weight.dtype)
            # 将原有的edge_weight和新生成的自环边的权重拼接起来
            self.edge_weight = torch.cat([self.edge_weight, self_loops_weight], dim=0)

        self.n_node_features = self.x.size(1)
        self.size_factors = self.x.sum(1)  # fix for ATAC case

    def __len__(self):
        """Return the number of observations stored in SpatialAnnTorchDataset"""
        return self.x.size(0)


def sparse_mx_to_sparse_tensor(sparse_mx: csr_matrix) -> SparseTensor:
    """
    Convert a scipy sparse matrix into a torch_sparse SparseTensor.

    Parameters
    ----------
    sparse_mx:
        Sparse scipy csr_matrix.

    Returns
    ----------
    sparse_tensor:
        torch_sparse SparseTensor object that can be utilized by PyG.
    """
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    indices = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)
    torch_sparse_coo_tensor = torch.sparse.FloatTensor(indices, values, shape)
    sparse_tensor = SparseTensor.from_torch_sparse_coo_tensor(
        torch_sparse_coo_tensor)
    return sparse_tensor


"""
This module contains data readers for the training of Garfield model.
"""

import os
import numpy as np
import pandas as pd
import scipy
from scipy.sparse import issparse, csr
from glob import glob

from anndata import AnnData
from muon import MuData
import muon as mu
import scanpy as sc


def read_mtx(path):
    """\
    Read mtx format data folder including:

        * matrix file: e.g. count.mtx or matrix.mtx or their gz format
        * barcode file: e.g. barcode.txt
        * feature file: e.g. feature.txt

    Parameters
    ----------
    path
        the path store the mtx files

    Return
    ----------
    AnnData
    """
    for filename in glob(path + '/*'):
        if ('count' in filename or 'matrix' in filename or 'data' in filename) and ('mtx' in filename):
            adata = sc.read_mtx(filename).T
    for filename in glob(path + '/*'):
        if 'barcode' in filename:
            barcode = pd.read_csv(filename, sep='\t', header=None).iloc[:, -1].values
            adata.obs = pd.DataFrame(index=barcode)
        if 'gene' in filename or 'peaks' in filename:
            gene = pd.read_csv(filename, sep='\t', header=None).iloc[:, -1].values
            adata.var = pd.DataFrame(index=gene)
        elif 'feature' in filename:
            gene = pd.read_csv(filename, sep='\t', header=None).iloc[:, 1].values
            adata.var = pd.DataFrame(index=gene)
    return adata

def read_scData(path, backed=False):
    """
    Read single cell dataset from single file

    Parameters
    ----------
    path
        the path store the file

    Return
    ------
    AnnData
    """
    if path is not None:
        path = path
    else:
        path = os.getcwd()

    if os.path.exists(path + '.h5ad'): # path=file name
        adata = sc.read_h5ad(path + '.h5ad', backed=backed)
    elif os.path.isdir(path):  # mtx format
        adata = read_mtx(path)
    elif os.path.isfile(path):
        if path.endswith(('.csv', '.csv.gz')):
            adata = sc.read_csv(path).T
        elif path.endswith(('.txt', '.txt.gz', '.tsv', '.tsv.gz')):
            df = pd.read_csv(path, sep='\t', index_col=0).T
            adata = AnnData(df.values, dict(obs_names=df.index.values), dict(var_names=df.columns.values))
        elif path.endswith('.h5ad'):
            adata = sc.read_h5ad(path)
        elif path.endswith('.loom'):
            adata = sc.read_loom(path, sparse=False)
        elif path.endswith('.h5mu'):
            adata = mu.read(path, backed=backed)
        elif path.endswith(tuple(['.h5mu/rna', '.h5mu/atac'])):
            adata = mu.read(path, backed=backed)
    else:
        raise ValueError("File {} not exists".format(path))

    if not issparse(adata.X) and not backed:
        adata.X = scipy.sparse.csr_matrix(adata.X)
    adata.var_names_make_unique()
    return adata

def read_multi_scData(root):
    """
    Read single cell dataset from multiple files

    Parameters
    ----------
    root
        the root store the single-cell data files, each file represent one dataset

    Return
    ----------
    AnnData
    """
    if root.split('/')[-1] == '*':
        adata = []
        for root in sorted(glob(root)):
            adata.append(read_scData(root))
        return AnnData.concatenate(*adata, batch_key='sub_batch', index_unique=None)
    else:
        return read_scData(root)

def concat_data(
        data_list,
        batch_categories=None,
        join='inner',
        batch_key='batch',
        index_unique=None,
        save=None
):
    """
    Concatenate multiple datasets along the observations axis with name ``batch_key``.

    Parameters
    ----------
    data_list
        A path list of AnnData matrices to concatenate with. Each matrix is referred to as a “batch”.
    batch_categories
        Categories for the batch annotation. By default, use increasing numbers.
    join
        Use intersection ('inner') or union ('outer') of variables of different batches. Default: 'inner'.
    batch_key
        Add the batch annotation to obs using this key. Default: 'batch'.
    index_unique
        Make the index unique by joining the existing index names with the batch category, using index_unique='-', for instance. Provide None to keep existing indices.
    save
        Path to save the new merged AnnData. Default: None.

    Returns
    ----------
    New merged AnnData.
    """
    if len(data_list) == 1:
        return read_multi_scData(data_list[0])
    elif isinstance(data_list, str):
        return read_multi_scData(data_list)
    elif isinstance(data_list, (AnnData, MuData)):
        return data_list

    adata_list = []
    for root in data_list:
        if isinstance(root, (AnnData, MuData)):
            adata = root
        else:
            adata = read_multi_scData(root)
        adata_list.append(adata)

    if isinstance(data_list[0], AnnData):
        if batch_categories is None:
            batch_categories = list(map(str, range(len(adata_list))))
        else:
            assert len(adata_list) == len(batch_categories)
        # [print(b, adata.shape) for adata,b in zip(adata_list, batch_categories)]
        concat = AnnData.concatenate(*adata_list, join=join, batch_key=batch_key,
                                     batch_categories=batch_categories, index_unique=index_unique)
    elif isinstance(data_list[0], MuData):
        concat = adata_list ## MuData

    if batch_key not in concat.obs:
        concat.obs[batch_key] = 'batch'
    concat.obs[batch_key] = concat.obs[batch_key].astype('category')

    if save:
        concat.write(save, compression='gzip')
    return concat






"""
Functions for graph construction
"""
import numpy as np
import warnings
import scanpy as sc
import scipy.sparse as sp
import igraph as ig
import leidenalg
import pynndescent

from . import _utils as utils

def get_nearest_neighbors(query_arr, target_arr, svd_components=None, randomized_svd=False, svd_runs=1,
                          metric='correlation'):
    """
    For each row in query_arr, compute its nearest neighbor in target_arr.

    Parameters
    ----------
    query_arr: np.array of shape (n_samples1, n_features)
        The query data matrix.
    target_arr: np.array of shape (n_samples2, n_features)
        The target data matrix.
    svd_components: None or int, default=None
        If not None, will first conduct SVD to reduce the dimension
        of the vertically stacked version of query_arr and target_arr.
    randomized_svd: bool, default=False
        Whether to use randomized SVD.
    svd_runs: int, default=1
        Run multiple instances of SVD and select the one with the lowest Frobenious reconstruction error.
    metric: string, default='correlation'
        The metric to use in nearest neighbor search.

    Returns
    -------
    neighbors: np.array of shape (n_samples1)
        The i-th element is the index in target_arr to whom the i-th row of query_arr is closest to.
    dists: np.array of shape (n_samples1)
        The i-th element is the distance corresponding to neighbors[i].
    """
    query_arr = utils.convert_to_numpy(query_arr)
    target_arr = utils.convert_to_numpy(target_arr)
    arr = np.vstack([query_arr, target_arr])
    arr = utils.svd_embedding(
        arr=arr, n_components=svd_components,
        randomized=randomized_svd,
        n_runs=svd_runs
    )
    query_arr = arr[:query_arr.shape[0], :]
    pivot_arr = arr[query_arr.shape[0]:, :]
    # approximate nearest neighbor search
    index = pynndescent.NNDescent(pivot_arr, n_neighbors=100, metric=metric)
    neighbors, dists = index.query(query_arr, k=50)
    neighbors, dists = neighbors[:, 0], dists[:, 0]
    return neighbors, dists



"""
This module contains generic base model functionalities, added as a Mixin to the
NicheCompass model.
"""

import inspect
import os
from copy import deepcopy
import warnings
from typing import Optional

import numpy as np
# import pickle
import scipy.sparse as sp
import torch
from anndata import AnnData

from .utils import load_saved_files, save_model_with_fallback


class BaseModelMixin():
    """
    Base model mix in class for universal model functionalities.

    Parts of the implementation are adapted from
    https://github.com/theislab/scarches/blob/master/scarches/models/base/_base.py#L15
    (01.10.2022) and
    https://github.com/scverse/scvi-tools/blob/master/scvi/model/base/_base_model.py#L63
    (01.10.2022).
    """

    def _get_user_attributes(self):
        # returns all the self attributes defined in a model class, eg, self.is_trained_
        attributes = inspect.getmembers(self, lambda a: not (inspect.isroutine(a)))
        attributes = [
            a for a in attributes if not (a[0].startswith("__") and a[0].endswith("__"))
        ]
        attributes = [a for a in attributes if not a[0].startswith("_abc_")]
        return attributes

    def _get_public_attributes(self):
        public_attributes = self._get_user_attributes()
        public_attributes = {a[0]: a[1] for a in public_attributes if a[0][-1] == "_"}
        return public_attributes

    def save(self,
             dir_path: str,
             overwrite: bool = False,
             save_adata: bool = False,
             adata_file_name: str = "adata_ref.h5ad",
             **anndata_write_kwargs):
        """
        Save model to disk (the Trainer optimizer state is not saved).

        Parameters
        ----------
        dir_path:
            Path of the directory where the model will be saved.
        overwrite:
            If `True`, overwrite existing data. If `False` and directory
            already exists at `dir_path`, error will be raised.
        save_adata:
            If `True`, also saves the AnnData object.
        adata_file_name:
            File name under which the AnnData object will be saved.
        adata_write_kwargs:
            Kwargs for adata write function.
        """
        if not os.path.exists(dir_path) or overwrite:
            os.makedirs(dir_path, exist_ok=overwrite)
        else:
            raise ValueError(f"Directory '{dir_path}' already exists."
                             "Please provide another directory for saving.")

        model_save_path = os.path.join(dir_path, "model_params.pt")
        attr_save_path = os.path.join(dir_path, "attr.pkl")
        var_names_save_path = os.path.join(dir_path, "var_names.csv")

        if save_adata:
            # Convert storage format of adjacency matrix to be writable by
            # adata.write()
            if "spatial_connectivities" in self.adata.obsp.keys() and self.adata.obsp["spatial_connectivities"] is not None:
                self.adata.obsp["spatial_connectivities"] = sp.csr_matrix(
                    self.adata.obsp["spatial_connectivities"])
            if "connectivities" in self.adata.obsp.keys() and self.adata.obsp["connectivities"] is not None:
                self.adata.obsp["connectivities"] = sp.csr_matrix(
                    self.adata.obsp["connectivities"])
            self.adata.write(
                os.path.join(dir_path, adata_file_name), **anndata_write_kwargs)

        var_names = self.adata.var_names.astype(str).to_numpy()
        public_attributes = self._get_public_attributes()

        torch.save(self.model.state_dict(), model_save_path)
        np.savetxt(var_names_save_path, var_names, fmt="%s")
        save_model_with_fallback(public_attributes, attr_save_path)
        # with open(attr_save_path, "wb") as f:
        #     pickle.dump(public_attributes, f)

    @classmethod
    def load_query_data(cls,
             dir_path: str,
             query_adata: Optional[AnnData] = None,
             ref_adata_name: str = "adata_ref.h5ad",
             use_cuda: bool = True,
             unfreeze_all_weights: bool = False,
             unfreeze_eps_weight: bool = False,
             unfreeze_layer0: bool = False,
             **kwargs
             ) -> torch.nn.Module:
        """
        Instantiate a model from saved output. Can be used for transfer learning
        scenarios and to learn de-novo gene programs by adding add-on gene
        programs and freezing non add-on weights.

        Parameters
        ----------
        dir_path:
            Path to saved outputs.
        adata:
            AnnData organized in the same way as data used to train the model.
            If ´None´, will check for and load adata saved with the model.
        adata_file_name:
            File name of the AnnData object to be loaded.
        use_cuda:
            If `True`, load model on GPU.
        unfreeze_all_weights:
            If `True`, unfreeze all weights.

        Returns
        -------
        model:
            Model with loaded state dictionaries and, if specified, frozen non
            add-on weights.
        """
        use_cuda = use_cuda and torch.cuda.is_available()
        map_location = torch.device("cpu") if use_cuda is False else None

        model_state_dict, var_names, attr_dict, adata_concat = (
            load_saved_files(dir_path=dir_path,
                             query_adata=query_adata,
                             ref_adata_name=ref_adata_name,
                             map_location=map_location))
        # print('model_state_dict.keys() is', model_state_dict.keys())

        init_params = deepcopy(cls._get_init_params_from_dict(attr_dict))
        # don't preprocess the data
        init_params['min_features'] = 0
        init_params['adata_list'] = adata_concat
        init_params.update(kwargs)

        # model = initialize_model(cls, init_params)
        model = cls(init_params)

        # set saved attrs for loaded model
        for attr, val in init_params.items():
            setattr(model, attr, val)

        # TODO 可能缺少了 batch normalization的层参数，当 used_DSBN=False 时
        model.model.load_state_dict(model_state_dict, strict=False)

        if use_cuda:
            model.model.cuda()
        model.model.eval()

        # First freeze all parameters and then subsequently unfreeze based on
        # load settings
        for param_name, param in model.model.named_parameters():
            param.requires_grad = False
            model.freeze_ = True
        if unfreeze_all_weights:
            for param_name, param in model.model.named_parameters():
                param.requires_grad = True
            model.freeze_ = False
        if unfreeze_eps_weight:
            # allow updates of eps_weight
            for param_name, param in model.model.named_parameters():
                if "eps_weight" in param_name or \
                        "eps_bias" in param_name:
                    param.requires_grad = True
        if unfreeze_layer0:
            # Allow updates of the first embedder weights
            for param_name, param in model.model.named_parameters():
                if ("layers.0" in param_name) or ("norm.0" in param_name):
                    param.requires_grad = True

        return model

    def _check_if_trained(self,
                          warn: bool = True):
        """
        Check if the model is trained.

        Parameters
        ----------
        warn:
             If not trained and `warn` is True, raise a warning, else raise a
             RuntimeError.
        """
        message = ("Trying to query inferred values from an untrained model. "
                   "Please train the model first.")
        if not self.is_trained_:
            if warn:
                warnings.warn(message)
            else:
                raise RuntimeError(message)


"""Preprocessing"""
from ._utils import (
    get_centroids,
    summarize_clustering,
    drop_zero_variability_columns,
    robust_svd,
    svd_embedding,
    tfidf,
    GeneScores,
    gene_scores
)

from ._graph import (
    get_nearest_neighbors
)

from .preprocess_utils import (
    preprocessing_rna,
    preprocessing_atac,
    preprocessing_adt,
    preprocessing # TODO
)
from .preprocess import DataProcess